import os, re, datetime, getpass, logging, sys
import numpy as np
import subprocess as sp
import matplotlib.pyplot as plt

from scipy import interpolate, constants
from dataclasses import dataclass
from matplotlib.ticker import FuncFormatter
from matplotlib.animation import FuncAnimation
from matplotlib.backends.backend_pdf import PdfPages, Stream, PdfFile
from fiqus.utils.Utils import FilesAndFolders


@dataclass
class PlotData:
    x_data: np.ndarray
    y_data: np.ndarray
    label: str
    color: np.ndarray
    linestyle: str = '-'


class PDFreport:
    def __init__(self, filepath, title='PostprocReport'):
        self.creation_date = datetime.datetime.now()
        self.author = getpass.getuser()
        self.file = os.path.join(filepath, str(title) + '.pdf')
        self.pdf = PdfPages(self.file)
        # init metadata
        self.add_metadata()

    def multi_sim_report(self):
        sims = self.plotter.simulation_collection

    def add_table(self, title, data, row_header=None, col_header=None):
        title_text = title
        footer_text = self.author + ' - ' + str(datetime.datetime.now())
        fig_background_color = 'white'
        fig_border = 'steelblue'

        column_headers = ['taus (s)', 'kappas (A)', 'alphas']
        row_headers = []
        # while I'm at it.
        cell_text = []
        for idx, row in enumerate(data):
            cell_text.append([f'{x:f}' for x in row])
            row_headers.append('k=' + str(idx + 1))
        # Get some lists of color specs for row and column headers
        rcolors = plt.cm.BuPu(np.full(len(row_headers), 0.1))
        ccolors = plt.cm.BuPu(np.full(len(column_headers), 0.1))  # Create the figure. Setting a small pad on tight_layout
        # seems to better regulate white space. Sometimes experimenting
        # with an explicit figsize here can produce better outcome.
        plt.figure(linewidth=2, edgecolor=fig_border, facecolor=fig_background_color, tight_layout={'pad': 1}, figsize=(5, 3))
        the_table = plt.table(cellText=cell_text, rowLabels=row_headers, rowColours=rcolors, rowLoc='right',
                              colColours=ccolors, colLabels=column_headers, loc='center')
        # Scaling is the only influence we have over top and bottom cell padding.
        # Make the rows taller (i.e., make cell y scale larger).
        the_table.scale(1, 1.5)  # Hide axes
        ax = plt.gca()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)  # Hide axes border
        plt.box(on=None)  # Add title
        plt.suptitle(title_text)  # Add footer
        plt.figtext(0.95, 0.05, footer_text, horizontalalignment='right', size=6, weight='light')  # Force the figure to update, so backends center objects correctly within the figure.
        # Without plt.draw() here, the title will center on the axes and not the figure.
        plt.draw()  # Create image. plt.savefig ignores figure edge and face colors, so map them.
        fig = plt.gcf()
        self.savefigs()
        plt.close()

    def savefigs(self):
        """ Save all open figures one on each page """
        for i in plt.get_fignums():
            self.pdf.savefig(i)

    def add_metadata(self):
        d = self.pdf.infodict()
        d['Title'] = 'Report'
        d['Author'] = getpass.getuser()
        d['CreationDate'] = datetime.datetime.now()

    def __del__(self):
        """ save all generated figures and close report on destruction """
        self.savefigs()
        self.pdf.close()


class PlotPythonConductorAC:
    """
    This class handles various plots for single simulation CACStrand models.
    """

    def __init__(self, fdm, outputs_folder_path='', simulation_data=None, flux_model=None) -> None:
        self.fdm = fdm
        self.outputs_folder_path = os.path.join(outputs_folder_path, fdm.magnet.postproc.output_folder)  # This is the path to the folder where the postprocessed data is written
        self.simulation_data = simulation_data
        self.flux_model = flux_model
        # self.pdf_report = PDFreport(fdm=fdm,author=getpass.getuser(),outputs_folder_path=self.simulation_data.solution_folder) if fdm.magnet.postproc.generate_pdf else None

    def colors(self, length):
        """ unify color scheme for all plots """
        return plt.cm.rainbow(np.linspace(0.05, 0.9, length))

    def export2txt(self, data_names, data_columns, filename):
        """ This function can be used to export one or more arrays of simulation data in one txt file (stored in solution folder)."""
        np_data = np.array(data_columns).T
        np.savetxt(os.path.join(self.simulation_data.solution_folder, str(filename) + '.txt'), np_data, header=" ".join(data_names), comments="")

    def plot_power_loss(self):
        """ Compares the losses produced by the ROHF model S_chain to the corresponding filament losses """

        t_final = np.array(self.simulation_data.time)[-1]
        if self.flux_model:
            _, p_irr_rohf, p_eddy_rohf, per_cycle_rohf, _ = self.flux_model.compute(self.simulation_data.time, self.simulation_data.I_transport, sim=self.simulation_data)
        if self.simulation_data.f and 1 / self.simulation_data.f < t_final:
            offset_time = t_final - 1 / self.simulation_data.f
            offset_idx = np.abs(self.simulation_data.time - offset_time).argmin()
            lastp_time = self.simulation_data.time[offset_idx:]
            lastp_I = self.simulation_data.I_transport[offset_idx:]
            lastp_int_flux = self.simulation_data.flux_internal[offset_idx:]
            lastp_filamentloss = self.simulation_data.power["FilamentLoss"][offset_idx:]
            # selffield_power_unitarea_loop_factor = 2 * selffield_power_unitarea # missing 2x radial contribution (eddy)
            print("#### HYSTERETIC SURFACE ####")
            print("simple flux surface:   " + str(np.trapz(lastp_I, lastp_int_flux)))
            print("###### PER CYCLE LOSS ######")
            print("FilamentLoss " + str(self.simulation_data.total_power_per_cycle["FilamentLoss"]))
            if self.flux_model: print("ROHF Pirr:        " + str(per_cycle_rohf))

        colorgrad = self.colors(3)
        #### DYNAMIC LOSS ####
        plt.figure()
        plt.title("Instantaneous power losses")
        plt.plot(self.simulation_data.time, self.simulation_data.power["TotalLoss"], color=colorgrad[0], linestyle='--', label="Total loss")
        plt.plot(self.simulation_data.time, self.simulation_data.power["FilamentLoss"], color=colorgrad[1], linestyle='--', label="Filament loss")
        plt.plot(self.simulation_data.time, self.simulation_data.power["EddyLoss"], color=colorgrad[2], linestyle='--', label="Eddy loss")
        # plt.plot(self.time, graphic_loss, color=self.c_grad[-2], label=r'graphic $P_{\text{irr}}$')
        if self.flux_model:
            plt.plot(self.simulation_data.time, p_irr_rohf + p_eddy_rohf, color=colorgrad[0], label=r'ROHF ($N=$' + str(self.flux_model.N) + r') $P_{\mathrm{total}}$')
            plt.plot(self.simulation_data.time, p_irr_rohf, color=colorgrad[1], label=r'ROHF ($N=$' + str(self.flux_model.N) + r') $P_{\mathrm{irr}}$')
            plt.plot(self.simulation_data.time, p_eddy_rohf, color=colorgrad[2], label=r'ROHF ($N=$' + str(self.flux_model.N) + r') $P_{\mathrm{eddy}}$')
            # plt.plot(self.simulation_data.time[:len(p_irr_rohf)], self.flux_model.Ec_0 * (abs(self.simulation_data.I_transport[:len(p_irr_rohf)])/self.flux_model.Ic_0)**(self.flux_model.n) * abs(self.simulation_data.I_transport[:len(p_irr_rohf)]), color=self.c_grad[-1], label = r'power-law Filament', linestyle='--')
        # plt.plot(self.time, abs(np.gradient(selffield_loss, self.time)) * (np.pi * strand_radius**2) , label=r'TEST selffield loss power')
        plt.xlabel(r'Time [s]')
        plt.ylabel(r"Power per unitlength [W']")
        plt.legend()
        plt.grid(linestyle='dotted')

    def plot_transport(self):
        """ This funcion makes a nice combined plot of Transport current and voltage """
        # make fancy plot

        colors = self.colors(2)
        fig, ax1 = plt.subplots()
        color = 'tab:red'
        ax1.set_title("Transport current and voltage")
        ax1.set_xlabel(r'Time $(s)$')
        ax1.set_ylabel(r'Current $(A)$', color=colors[0])
        ax1.plot(self.simulation_data.time, self.simulation_data.I_transport, color=colors[0])
        ax1.grid(axis='x', linestyle='dotted')
        ax1.tick_params(axis='y', labelcolor=colors[0])

        ax2 = ax1.twinx()
        color = 'tab:blue'
        ax2.set_ylabel(r'Voltage $(V)$', color=colors[1])
        ax2.plot(self.simulation_data.time, self.simulation_data.V_transport_unitlen, color=colors[1])
        ax2.tick_params(axis='y', labelcolor=colors[1])
        ax2.grid(axis='y', linestyle='dotted')
        fig.tight_layout()

    def plot_phaseshift_voltage(self, phaseshift):
        # removes phaseshift from voltage
        if not self.f:
            raise ValueError(f'No shift plot for selected source type')

        time_delay = phaseshift / (360 * self.f)
        shift_idx = (np.abs(self.simulation_data.time - time_delay)).argmin()
        shifted_V = self.simulation_data.V_transport[shift_idx:]
        shifted_time = self.simulation_data.time[1 + shift_idx:] - self.simulation_data.time[shift_idx]

        # make fancy plot
        fig, ax1 = plt.subplots()
        color = 'tab:red'
        ax1.set_title("shifted voltage (" + str(phaseshift) + " degree)")
        ax1.set_xlabel(r'Time $(s)$')
        ax1.set_ylabel(r'Current $(A)$', color=color)
        ax1.plot(self.simulation_data.time, self.simulation_data.I_transport, color=color)
        ax1.grid(axis='x', linestyle='dotted')
        ax1.tick_params(axis='y', labelcolor=color)

        ax2 = ax1.twinx()
        color = 'tab:blue'
        ax2.set_ylabel(r'Voltage $(V)$', color=color)
        ax2.plot(shifted_time, shifted_V, color=color, linestyle='dashed')
        ax2.tick_params(axis='y', labelcolor=color)
        ax2.grid(axis='y', linestyle='dotted')
        fig.tight_layout()

    def plot_flux(self):
        """ This plots the internal flux obtained through two approaches and compares them with an s_chain model """
        # L_base = (self.simulation_data.flux_internal[1]-self.simulation_data.flux_internal[0])/(self.simulation_data.I_transport[1]-self.simulation_data.I_transport[0])
        if self.flux_model:
            flux_rohf, _, _, _, _ = self.flux_model.compute(self.simulation_data.time, self.simulation_data.I_transport, sim=self.simulation_data)

        colors = self.colors(3)
        ### SIMPLE FLUX HOMOGENIZATION ###
        plt.figure()
        plt.title("Internal flux ")
        plt.xlabel(r"Transport current $[A]$")
        plt.ylabel(r"Flux per unitlength $[Tm]$")
        # plt.plot(self.I_transport, self.flux_external, color=self.c_grad[2], label = r'$\Phi_{\text{ext}}$')
        plt.plot(self.simulation_data.I_transport, self.simulation_data.flux_internal, color=colors[0], label=r'$\Phi_{\text{int}}$')
        plt.plot(self.simulation_data.I_transport, constants.mu_0 / (4 * np.pi) * self.simulation_data.I_transport, color=colors[1], label=r"$L'=\mu_0/(4 \pi)$")
        if self.flux_model: plt.plot(self.simulation_data.I_transport, flux_rohf, color=colors[2], label=r'$\Phi_{\text{ROHF}}$ ($N=$' + str(self.flux_model.N) + ')')
        plt.grid(linestyle='dotted')
        plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))
        plt.legend()
        plt.grid(axis='y', linestyle='dotted')

        ### FLUX BASED VOLTAGE ###
        if self.flux_model: voltage_rohf = -np.gradient((flux_rohf + self.simulation_data.flux_external), self.simulation_data.time)
        plt.figure()
        plt.title("flux based Voltage")
        plt.xlabel(r"Time $[s]$")
        plt.ylabel(r"Voltage per unit-length $\left[\frac{V}{m}\right]$")
        # plt.plot(self.time, -np.gradient((self.flux_internal + self.flux_external), self.time), color=self.c_grad[3], label = r'$-\frac{d\Phi}{dt}$ graphic')
        if self.flux_model: plt.plot(self.simulation_data.time, voltage_rohf, color=colors[2], label=r'$-\frac{d\Phi}{dt}$ ROHF ($N=$' + str(self.flux_model.N) + ')')
        plt.plot(self.simulation_data.time, self.simulation_data.V_transport_unitlen, color=colors[0], linestyle='--', label=r'$V$ transport')
        plt.grid(linestyle='dotted')
        plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        plt.legend()
        plt.grid(axis='y', linestyle='dotted')

        V = self.simulation_data.V_transport_unitlen
        I = self.simulation_data.I_transport
        idx = min(range(len(V)), key=lambda i: abs(V[i] + 1e-4))
        print("ciritical current: " + str(I[idx]))

        # data = np.vstack([self.simulation_data.I_transport, self.simulation_data.flux_internal, flux_rohf])
        # self.export2txt(['current','internal_flux','rohf_flux'], data, 'flux_hysteresis')

    def plot_current_rings(self):
        """
        This function decomposes the total transport current into an reversible and irreversible part based on inner flux computations of the filaments.
        It is basically the dual approach to the plot_voltage_decomp() function declared here after but shouldnt suffer from its numeric instabilities.
        """

        t_final = np.array(self.simulation_data.time)[-1]
        if not self.simulation_data.f or 1 / self.simulation_data.f > t_final:
            raise ValueError(r'Cant plot current rings')

        ### plot current threshold rings ###
        if self.flux_model:
            jc_real = 3.33e10
            circle_radii = np.sqrt(self.simulation_data.conductor.strand.number_of_filaments * (self.simulation_data.conductor.strand.filament_diameter / 2) ** 2 - np.array(self.flux_model.kappas_bar) / (jc_real * np.pi))
            circle_surfaces = np.pi * circle_radii ** 2
            ringwidth = np.diff(np.flip(circle_radii), prepend=0)

            plt.figure()
            plt.title("ROHF single filament - field free zones")
            for idx, radius in enumerate(circle_radii):
                shells = [plt.Circle((2.05 * circle_radii[0] * idx, 0), radius, color=self.c_grad[idx], hatch='//', label="> " + "{:.2f}".format(self.flux_model.kappas_bar[idx]) + " A", fill=False),
                          plt.Circle((2.05 * circle_radii[0] * idx, 0), circle_radii[0], color='black', fill=False)]  # contour
                for shell in shells: plt.gca().add_patch(shell)
            # core = plt.Circle((0,0), circle_radii[-1], color='white')
            # plt.gca().add_patch(core)
            plt.legend()
            plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))
            plt.xlim(-1.1 * circle_radii[0], 2.1 * len(circle_radii) * circle_radii[0])
            plt.ylim(-1.1 * circle_radii[0], 1.1 * circle_radii[0])

    def plot_fft_voltage(self):
        """ analyze the transport voltage in the frequency domain """
        if self.simulation_data.f:
            T = 1 / self.simulation_data.f
            start_idx = abs(self.simulation_data.time - (max(self.simulation_data.time) - T)).argmin() - 2
            X = np.fft.fft(self.simulation_data.V_transport_unitlen[start_idx:])
            N = len(X)
            n = np.arange(N)
            freq = n / T

            fig, axs = plt.subplots(2)
            fig.suptitle(r'Frequency domain voltage')
            axs[0].stem(freq, np.abs(X), 'b', markerfmt=" ", basefmt="-b")
            axs[0].set_xlabel('Freq (Hz)')
            axs[0].set_ylabel('FFT Amplitude |X(freq)|')
            axs[0].grid(linestyle='dotted')
            axs[0].set_xlim([0, 10 * self.simulation_data.f])

            axs[1].plot(self.simulation_data.time[:start_idx - 1], self.simulation_data.V_transport_unitlen[:start_idx - 1], '--r')
            axs[1].plot(self.simulation_data.time[start_idx:], np.fft.ifft(X), 'r', label='transport voltage')
            axs[1].set_xlabel('Time (s)')
            axs[1].set_ylabel('Amplitude')
            axs[1].grid(linestyle='dotted')
            axs[1].set_xlim([0, max(self.simulation_data.time)])

    def plot_instantaneous_power(self, show: bool = True, title: str = "Power", save_plot: bool = False, save_folder_path: str = None, save_file_name: str = None, overwrite: bool = False):
        plt.figure()
        plt.plot(self.simulation_data.power['Time'], self.simulation_data.power[self.simulation_data.power_columns[1:]], label=self.simulation_data.power_columns[1:])
        plt.xlabel('Time [s]')
        plt.ylabel('Power [W/m]')
        plt.legend()

        # Configure title:
        # Class attributes can be accessed by using '<< ... >>' in the title string.
        commands = re.findall('<<(.*?)>>', title)
        for c in commands:
            title = title.replace(f"<<{c}>>", str(eval('self.simulation_data.' + c)))
        plt.title(title)

        if save_plot:  # Save the plot
            filePath = os.path.join(save_folder_path, str(save_file_name) + '.png')
            plt.savefig(filePath)

        if show:
            plt.show()
        else:
            plt.close()

    def plot_impedance(self, frequency_range):
        """ This function compares the estimated data impedance with an analytic solution for inductive tranmission """

        if not self.simulation_data.f or not self.simulation_data.analytic_RL:
            raise ValueError(f'No impedance estimate for selected source type')
        analytic_impedance = self.simulation_data.analytic_RL[0] + 1j * 2 * np.pi * frequency_range * self.simulation_data.analytic_RL[1]  # R + jwL

        fig, axs = plt.subplots(2)
        fig.suptitle(r'Simulation Impedance $Z$')
        axs[0].plot(frequency_range, np.abs(np.imag(analytic_impedance)), '-b', label="pure Inductance")
        axs[0].plot(self.simulation_data.f, np.abs(self.simulation_data.data_impedance), 'r', marker="x", label="simulation")
        axs[0].set_xscale('log')
        axs[0].set_yscale('log')
        axs[0].set_ylabel(r'Magnitude $|Z|$')
        axs[0].grid(linestyle='dotted')
        axs[0].legend()

        axs[1].plot(frequency_range, 90 * np.ones((len(frequency_range), 1)), '-b', label="pure Inductance")
        axs[1].plot(self.simulation_data.f, np.angle(self.simulation_data.data_impedance, deg=True), 'r', marker="x", label="simulation")
        axs[1].set_xscale('log')
        axs[1].set_ylim([-15, 105])
        axs[1].set_ylabel(r'Phase $\measuredangle{Z}$' + r' $(deg)$')
        axs[1].set_xlabel(r'Frequency $(Hz)$')
        axs[1].grid(linestyle='dotted')
        axs[1].legend()

    def plot_resistivity(self):

        time, I, flux = self.simulation_data.first_rise(current=True)

        if self.flux_model:
            flux_rohf, p_irr, _, _, _ = self.flux_model.compute(time, I, sim=self.simulation_data)
            # p_ohm_fil, p_ohm_mat, R_fil, R_mat = self.flux_model.ohmic_loss(I, self.simulation_data)

        colors = self.colors(3)
        ### SIMPLE FLUX HOMOGENIZATION ###
        plt.figure(figsize=(5, 8))
        plt.title("Differential resitivity")
        plt.xlabel(r"Transport current $[A]$")
        plt.ylabel(r"Transport voltage per unit length $[V/m]$")
        # plt.plot(self.I_transport, self.flux_external, color=self.c_grad[2], label = r'$\Phi_{\text{ext}}$')
        plt.plot(I, abs(self.simulation_data.V_transport_unitlen[:len(I)]), color=colors[0], label=r'FEM reference')
        # plt.plot(I, (1.68e-10/(0.5*np.pi*(0.5e-3)**2))*I, color=self.c_grad[1], label='RRR100 copper')
        if self.flux_model:
            plt.plot(I, self.flux_model.Ec_0 * (abs(I) / self.flux_model.Ic_0) ** (self.flux_model.n), color=colors[1], label=r'Power-law only')
            plt.plot(I, (p_irr) / I, color=colors[2], label=r'ROHF current sharing')
            plt.axhline(y=self.flux_model.Ec_0, color='black', linestyle='--', label=r'$E_c = 1^{-4} \mathrm{V}/\mathrm{m}$')
            # plt.axis([0.7*self.flux_model.Ic_0, 1.2*self.flux_model.Ic_0, 0, 100*self.flux_model.Ec_0])
        plt.grid(linestyle='dotted')
        plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))
        plt.legend()
        plt.grid(axis='y', linestyle='dotted')

    def show(self):
        plt.show()


class BatchPlotPythonConductorAC:
    """
    This class loads and stores the data from the simulations specified in a csv file and can apply various postprocessing operations on the data.
    The data from each simulation is saved as a SimulationData object which is subsequently stored in a list in this class.
    """

    def __init__(self, fdm, lossMap_gridData_folder=None, outputs_folder_path='', simulation_collection=None, fluxmodel_collection=None) -> None:
        self.fdm = fdm
        self.outputs_folder_path = os.path.join(outputs_folder_path, fdm.magnet.postproc.output_folder)  # This is the path to the folder where the postprocessed data is written
        self.simulation_collection = simulation_collection
        self.fluxmodel_collection = fluxmodel_collection
        try:
            sp.run(['pdflatex', '--version'], stdout=sp.DEVNULL, stderr=sp.DEVNULL)
            tex_installation = True
        except:
            tex_installation = False
        self.tex_installation = tex_installation

        if not os.path.exists(self.outputs_folder_path):
            os.makedirs(self.outputs_folder_path)

        if lossMap_gridData_folder is not None:
            self.totalLoss_gridData = self.load_lossMap_gridData('TotalLoss', lossMap_gridData_folder)
            self.filamentLoss_gridData = self.load_lossMap_gridData('FilamentLoss', lossMap_gridData_folder)
            self.eddyLoss_gridData = self.load_lossMap_gridData('EddyLoss', lossMap_gridData_folder)
            self.couplingLoss_gridData = self.load_lossMap_gridData('CouplingLoss', lossMap_gridData_folder)

    def colors(self, length):
        """ unify color scheme for plots """
        return plt.cm.rainbow(np.linspace(0.05, 0.9, length))

    def create_non_overwriting_filepath(self, folder_path, base_name, extension, overwrite):
        """
            Creates a filepath that does not overwrite any existing files.

            This function checks if a file already exists at the specified filepath. If the file exists and `overwrite` is False,
            it modifies the filepath to create a new file instead of overwriting the existing one.
            If `overwrite` is True or the file does not exist, it returns the filepath as it is.

            Parameters
            ----------
            folder_path : str
                The path to the folder where the file will be created.
            base_name : str
                The base name of the file.
            extension : str
                The extension of the file.
            overwrite : bool, optional
                If True, the function will overwrite an existing file. If False, the function will modify the filepath to avoid overwriting. Defaults to False.

            Returns
            -------
            str
                The final filepath. If `overwrite` is False and a file already exists at the original filepath, this will be a new filepath that does not overwrite any existing files.
        """
        if os.path.exists(os.path.join(folder_path, base_name + extension)) and not overwrite:
            counter = 1
            new_name = base_name + f"_{counter}" + extension
            while os.path.exists(os.path.join(folder_path, new_name)):
                new_name = base_name + f"_{counter}" + extension
                counter += 1
            return os.path.join(folder_path, new_name)

        return os.path.join(folder_path, base_name + extension)

    def lossMap_createGridData(self, lossType='TotalLoss', x_val_to_include=None, y_val_to_include=None):
        """
        This function creates the grid data needed for the loss map, based on the yaml input file.
        Given a collection of simulations it interpolates the loss data between the datapoints to a grid and returns the grid data.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        # Extract data from simulation collection and normalize
        x_arr = np.array([eval('sd.' + lm.x_val) / lm.x_norm for sd in self.simulation_collection])
        y_arr = np.array([eval('sd.' + lm.y_val) / lm.y_norm for sd in self.simulation_collection])
        loss = np.array([sd.total_power_per_cycle[lossType] / lm.loss_norm for sd in self.simulation_collection])

        # Logarithmic scaling
        if lm.x_log: x_arr = np.log10(x_arr)
        if lm.y_log: y_arr = np.log10(y_arr)
        if lm.loss_log: loss = np.log10(loss)

        x_arr_interpolated = np.linspace(min(x_arr), max(x_arr), lm.x_steps)
        y_arr_interpolated = np.linspace(min(y_arr), max(y_arr), lm.y_steps)
        # Insert specific values to the grid if they are not already included (useful for cross sections)
        if x_val_to_include is not None and x_val_to_include not in x_arr_interpolated:
            x_arr_interpolated = np.insert(x_arr_interpolated, np.where(x_arr_interpolated > x_val_to_include)[0][0], x_val_to_include)
        if y_val_to_include is not None and y_val_to_include not in y_arr_interpolated:
            y_arr_interpolated = np.insert(y_arr_interpolated, np.where(y_arr_interpolated > y_val_to_include)[0][0], y_val_to_include)

        # Create grid
        X, Y = np.meshgrid(x_arr_interpolated, y_arr_interpolated, indexing='ij')
        gridPoints = np.c_[X.ravel(), Y.ravel()]
        dataPoints = np.c_[x_arr, y_arr]

        # Interpolate the simulation data onto the grid
        V = interpolate.griddata(
            dataPoints,
            loss,
            gridPoints,
            method='linear'  # Cubic produces cleaner plots. Any incentive to go back to linear?
        ).reshape(X.shape)

        return X, Y, V, dataPoints

    def save_lossMap_gridData(self, save_folder_name='lossMap_gridData'):
        """
        This function calls the lossMap_createGridData function and saves the grid data.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        lossTypes = ['TotalLoss', 'FilamentLoss', 'EddyLoss', 'CouplingLoss', 'CouplingLoss_dyn', 'TotalLoss_dyn']  # Only in the case dynamic correction is used, must be changed later
        # 1) Create a folder to store the output files
        gridData_folder_path = self.create_non_overwriting_filepath(self.outputs_folder_path, save_folder_name, '', self.fdm.run.overwrite)
        if not os.path.exists(gridData_folder_path): os.makedirs(gridData_folder_path)
        # 2) Create the grid data for each loss type and save it
        for lossType in lossTypes:
            X, Y, V, _ = self.lossMap_createGridData(lossType)
            if lm.x_log: X = np.power(10, X)
            if lm.y_log: Y = np.power(10, Y)
            if lm.loss_log: V = np.power(10, V)
            np.savetxt(os.path.join(gridData_folder_path, f'{lossType}.txt'), np.column_stack((X.ravel(), Y.ravel(), V.ravel())), delimiter=' ', header=f'{lm.x_val} {lm.y_val} {lossType}', comments='')

    def load_lossMap_gridData(self, lossType='TotalLoss', save_folder_name='lossMap_gridData'):
        """
        This function loads the grid data for a given loss type.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        gridData_folder_path = os.path.join(self.inputs_folder_path, save_folder_name)

        if not os.path.exists(gridData_folder_path):
            raise FileNotFoundError(f'The folder {gridData_folder_path} does not exist.')

        X, Y, V = np.loadtxt(os.path.join(gridData_folder_path, f'{lossType}.txt'), unpack=True, skiprows=1)

        if lm.x_log: X = np.log10(X)
        if lm.y_log: Y = np.log10(Y)
        if lm.loss_log: V = np.log10(V)

        # Get the unique counts of X and Y
        unique_X = np.unique(X)
        unique_Y = np.unique(Y)

        # Reshape the data
        X = X.reshape((len(unique_X), len(unique_Y)))
        Y = Y.reshape((len(unique_X), len(unique_Y)))
        V = V.reshape((len(unique_X), len(unique_Y)))

        return X, Y, V

    def save_magnetization(self):
        """
        This function saves the magnetization data for all simulations in the simulation collection.
        """
        magnetization_folder_path = self.create_non_overwriting_filepath(self.outputs_folder_path, 'magnetization', '', self.fdm.run.overwrite)
        if not os.path.exists(magnetization_folder_path): os.makedirs(magnetization_folder_path)
        for sd in self.simulation_collection:
            magnetization = sd.magn_fil + sd.magn_matrix
            magnetization = np.c_[sd.time, magnetization]
            np.savetxt(os.path.join(magnetization_folder_path, f'magn_f{sd.solve.source_parameters.sine.frequency}_b{sd.solve.source_parameters.sine.field_amplitude}_I{sd.solve.source_parameters.sine.current_amplitude}.txt'), magnetization, delimiter=' ', header='t x y z', comments='')

    def lossMap_crossSection(self, slice_value, axis_to_cut='x'):
        """
        This function returns the data corresponding to a cross section of the loss map, for all loss types.
        Given an axis and a value, it sweeps the other axis for the closest value and returns the data.
        Example: Given slice value 0 and axis x, it returns the data for the cross section at x = 0.
        """

        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        if axis_to_cut == 'x':
            x_val_to_include = slice_value
            y_val_to_include = None
        elif axis_to_cut == 'y':
            x_val_to_include = None
            y_val_to_include = slice_value
        X, Y, V, dataPoints = self.lossMap_createGridData('TotalLoss', x_val_to_include, y_val_to_include)
        _, _, FilamentLoss, _ = self.lossMap_createGridData('FilamentLoss', x_val_to_include, y_val_to_include)
        _, _, EddyLoss, _ = self.lossMap_createGridData('EddyLoss', x_val_to_include, y_val_to_include)
        _, _, CouplingLoss, _ = self.lossMap_createGridData('CouplingLoss', x_val_to_include, y_val_to_include)

        if axis_to_cut == 'x':
            index = np.abs(X[:, 0] - slice_value).argmin()
            slice_vals = Y[index, :]

        elif axis_to_cut == 'y':
            index = np.abs(Y[0, :] - slice_value).argmin()
            slice_vals = X[:, index]

        # Extract the loss values for the constant frequency across all applied fields
        totalLoss = V[index, :] if axis_to_cut == 'x' else V[:, index]
        filamentLoss = FilamentLoss[index, :] if axis_to_cut == 'x' else FilamentLoss[:, index]
        eddyLoss = EddyLoss[index, :] if axis_to_cut == 'x' else EddyLoss[:, index]
        couplingLoss = CouplingLoss[index, :] if axis_to_cut == 'x' else CouplingLoss[:, index]

        return slice_vals, totalLoss, filamentLoss, eddyLoss, couplingLoss

    def plot_lossMap_crossSection(self):
        """
        This function calls the lossMap_crossSection function and plots the data it returns, which is the loss for all values of one axis, given a constant value of the other axis.
        """

        if self.tex_installation:
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            plt.rcParams['font.size'] = 20

        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        slice_value = lm.cross_section.cut_value
        axis_to_cut = lm.cross_section.axis_to_cut

        if (lm.x_log and axis_to_cut == 'x') or (lm.y_log and axis_to_cut == 'y'):
            slice_value = np.log10(slice_value)

        slice_vals, totalLoss, filamentLoss, eddyLoss, couplingLoss = self.lossMap_crossSection(slice_value, axis_to_cut=axis_to_cut)

        def log_formatter(x, pos):
            """
                Format the tick labels on the plot.
            """
            return f"$10^{{{int(x)}}}$"

        # Plot the loss with respect to applied field for the constant frequency
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.plot(slice_vals, totalLoss, label=f'Total Loss')
        ax.plot(slice_vals, filamentLoss, label=f'Filament Loss')
        ax.plot(slice_vals, eddyLoss, label=f'Eddy Loss')
        ax.plot(slice_vals, couplingLoss, label=f'Coupling Loss')

        tick_formatter = FuncFormatter(log_formatter)
        if lm.x_log and axis_to_cut == 'y' or lm.y_log and axis_to_cut == 'x':
            ax.xaxis.set_major_formatter(tick_formatter)
        if lm.loss_log:
            ax.yaxis.set_major_formatter(tick_formatter)

        title = lm.cross_section.title.replace('<<cut_value>>', str(round(10 ** slice_value, 3)))
        ax.set_title(title)
        ax.set_xlabel(lm.ylabel if axis_to_cut == 'x' else lm.xlabel)
        ax.set_ylabel(lm.cross_section.ylabel)
        ax.legend()

        # np.savetxt(os.path.join(self.outputs_folder_path, 'lossMaps_cut_0p2T_0A.txt'), np.column_stack((10**slice_vals, 10**totalLoss, 10**eddyLoss, 10**couplingLoss, 10**filamentLoss)), delimiter=' ', header='f total eddy coupling filament', comments='')

        if lm.cross_section.save_plot:
            filePath = self.create_non_overwriting_filepath(self.outputs_folder_path, lm.cross_section.filename, '.png', self.fdm.run.overwrite)
            plt.savefig(filePath)

        if self.fdm.run.launch_gui: plt.show()

    def plot_percycle_losses(self, sims, flux_model, title):
        """ This function produces a frequency to per cycle power loss plot used within the ROHF parameter optimization """

        pc_irr_rohf = []
        pc_eddy_rohf = []
        pc_fil_sim = [sim.total_power_per_cycle['FilamentLoss'] for sim in sims]
        pc_eddy_sim = [sim.total_power_per_cycle['EddyLoss'] for sim in sims]
        freq = [sim.f for sim in sims]

        for sim in sims:
            _, _, _, irr, eddy = flux_model.compute(time=sim.time, I=sim.I_transport, sim=sim)
            pc_irr_rohf.append(irr)
            pc_eddy_rohf.append(eddy)

        plt.figure()
        plt.title(title)
        plt.plot(freq, pc_fil_sim, 'o-', label='FilamentLoss')
        plt.plot(freq, pc_irr_rohf, 'x-', label=r'$P_{\mathrm{irr}}$ ROHF')
        plt.plot(freq, pc_eddy_sim, 'o-', label='EddyLoss')
        plt.plot(freq, pc_eddy_rohf, 'x-', label=r'$P_{\mathrm{eddy}}$ ROHF')
        plt.legend()
        plt.xlabel(r'Frequency ($\mathrm{Hz}$)')
        plt.ylabel(r'Per cycle power loss ($\mathrm{W/m}$)')
        plt.xscale('log')
        plt.yscale('log')

    def plot_flux_time_triplet(self, flux_model, title):

        amplitudes = sorted(list(set([sim.I_amp for sim in self.simulation_collection])))

        high_I_sims = []
        mid_I_sims = []
        low_I_sims = []
        for sim in self.simulation_collection:
            if sim.I_amp == amplitudes[-1]:
                high_I_sims.append(sim)
            elif sim.I_amp == amplitudes[-2]:
                mid_I_sims.append(sim)
            elif sim.I_amp == amplitudes[2]:
                low_I_sims.append(sim)

        colors = self.colors(max([len(high_I_sims), len(mid_I_sims), len(low_I_sims)]))
        plt.figure()
        plt.title(title)
        plt.subplot(3, 1, 1)
        for i, sim in enumerate(high_I_sims):
            plt.plot(sim.time * sim.f, sim.flux_internal, color=colors[i], linestyle='--')
            plt.plot(sim.time * sim.f, flux_model.compute(I=sim.I_transport, time=sim.time, sim=sim)[0], color=colors[i], label=str(sim.f) + ' Hz')
        plt.legend()

        plt.subplot(3, 1, 2)
        for i, sim in enumerate(mid_I_sims):
            plt.plot(sim.time * sim.f, sim.flux_internal, color=colors[i], linestyle='--')
            plt.plot(sim.time * sim.f, flux_model.compute(I=sim.I_transport, time=sim.time, sim=sim)[0], color=colors[i], label=str(sim.f) + ' Hz')

        plt.subplot(3, 1, 3)
        for i, sim in enumerate(low_I_sims):
            plt.plot(sim.time * sim.f, sim.flux_internal, color=colors[i], linestyle='--')
            plt.plot(sim.time * sim.f, flux_model.compute(I=sim.I_transport, time=sim.time, sim=sim)[0], color=colors[i], label=str(sim.f) + ' Hz')
            # plt.ylabel(r'Internal Flux per unit length ($\mathrm{Tm}$)')
        plt.text(0.04, 0.5, r'Internal Flux per unit length ($\mathrm{Tm}$)', va='center', rotation='vertical')
        plt.xlabel(r'Time in periods')

    def animate_lossMap_crossSection(self):
        """
        This function is similar to the plot_lossMap_crossSection function, but instead of plotting the loss for at a constant crossection,
        it sweeps the crossection over a chosen axis and plots the loss for each crossection as an animation.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        axis = lm.cross_section_sweep.axis_to_sweep

        X, Y, V, dataPoints = self.lossMap_createGridData('TotalLoss')
        x_vals = X[:, 0]  # x-values from the loss map
        y_vals = Y[0, :]  # y-values from the loss map

        if axis == 'x':
            A = np.zeros((lm.y_steps, 4, lm.x_steps))
            axis_to_sweep = x_vals
            constant_axis = y_vals
        elif axis == 'y':
            A = np.zeros((lm.x_steps, 4, lm.y_steps))
            axis_to_sweep = y_vals
            constant_axis = x_vals

        for i, val in enumerate(axis_to_sweep):
            _, totalLoss, filamentLoss, eddyLoss, couplingLoss = self.lossMap_crossSection(val, axis_to_cut=axis)
            A[:, 0, i] = totalLoss
            A[:, 1, i] = filamentLoss
            A[:, 2, i] = eddyLoss
            A[:, 3, i] = couplingLoss

        # Initialize the plot
        fig, ax = plt.subplots()
        lines = ax.plot(constant_axis, A[:, :, 0], lw=2, label=['total Loss', 'filament Loss', 'eddy Loss', 'coupling Loss'])

        # Set plot limits and labels
        ax.set_xlim(constant_axis[0], constant_axis[-1])
        ax.set_ylim(np.min(A), np.max(A))
        ax.set_xlabel(lm.ylabel if axis == 'x' else lm.xlabel)
        ax.set_ylabel(lm.cross_section_sweep.ylabel)

        # Define the animation update function
        def update(frame):
            for i, line in enumerate(lines):
                line.set_ydata(A[:, i, frame])

            if axis == 'x':
                if lm.x_log:
                    sweep_value = 10 ** x_vals[frame]
                else:
                    sweep_value = x_vals[frame]
            elif axis == 'y':
                if lm.y_log:
                    sweep_value = 10 ** y_vals[frame]
                else:
                    sweep_value = y_vals[frame]

            title = lm.cross_section_sweep.title.replace('<<sweep_value>>', str(round(sweep_value, 3)))
            ax.set_title(title)
            return lines,  # line1, line2, line3, line4

        # Create the animation
        dt = 0.1
        ani = FuncAnimation(fig, update, frames=lm.x_steps if axis == 'x' else lm.y_steps, interval=dt * 1000, blit=False)

        # Show the animation
        plt.legend()
        plt.grid()

        if lm.cross_section_sweep.save_plot:
            filepath = self.create_non_overwriting_filepath(folder_path=self.outputs_folder_path, base_name=lm.cross_section_sweep.filename, extension='.gif', overwrite=self.fdm.run.overwrite)
            ani.save(filepath, writer='imagemagick', fps=1 / dt)

        if self.fdm.run.launch_gui: plt.show()

    def create_lossMap(self):
        """
        This function creates a loss map based on the inputs given in the loss_map section of the input file.
        The loss-map can be plotted and saved as a .png file.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        if self.simulation_collection:
            X, Y, V, dataPoints = self.lossMap_createGridData(lm.loss_type)
        else:
            X, Y, V = self.totalLoss_gridData

        if self.tex_installation:
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            plt.rcParams['font.size'] = 20

        fig, ax = plt.subplots(figsize=(10, 8))

        c = plt.pcolormesh(X, Y, V, shading='gouraud', cmap='plasma_r')
        c_min = min([np.ceil(np.min(V)) for V in [V]])
        c_max = max([np.floor(np.max(V)) for V in [V]])
        c_ticks = [int(val) for val in np.arange(c_min, c_max + 1)]
        cont = plt.contour(X, Y, V, c_ticks, colors='k', linestyles='dashed')

        if lm.show_datapoints:
            plt.scatter(dataPoints[:, 0], dataPoints[:, 1], s=50, edgecolors='k')

        if lm.show_loss_type_dominance_contour:
            sigmoid = lambda x: 1 / (1 + np.exp(-x))
            if self.simulation_collection:
                _, _, FilamentLoss, _ = self.lossMap_createGridData(lossType='FilamentLoss')
                _, _, CouplingLoss, _ = self.lossMap_createGridData(lossType='CouplingLoss')
                _, _, EddyLoss, _ = self.lossMap_createGridData(lossType='EddyLoss')
            # else:
            #     _, _, FilamentLoss = self.filamentLoss_gridData
            #     _, _, CouplingLoss = self.couplingLoss_gridData
            #     _, _, EddyLoss = self.eddyLoss_gridData
            fil_vs_coupling_loss = np.maximum(FilamentLoss, EddyLoss) - CouplingLoss
            fil_vs_eddy_loss = EddyLoss - np.maximum(FilamentLoss, CouplingLoss)
            plt.contour(X, Y, sigmoid(fil_vs_coupling_loss), [0.5], colors='k')
            plt.contour(X, Y, sigmoid(fil_vs_eddy_loss), [0.5], colors='k')

        cbar = fig.colorbar(c, ticks=c_ticks)  # , labels=c_labels)
        # cbar.ax.set_xticks([-7, -6, -5, -4, -3, -2, -1, 0, 1])
        # cbar.ax.set_yticklabels([r'$10^{-7}$', r'$10^{-6}$', r'$10^{-5}$', r'$10^{-4}$', r'$10^{-3}$', r'$10^{-2}$', r'$10^{-1}$', r'$10^0$', r'$10^1$'])
        cbar.ax.set_yticklabels([f"$10^{{{val}}}$" for val in c_ticks])
        # plt.grid(alpha=0.5)
        # plt.title(lm.title)
        # plt.xlabel(lm.xlabel)
        # plt.ylabel(lm.ylabel)
        plt.title(r'Loss per cycle (J/m)')
        plt.xlabel(r'Frequency $f$ (Hz)')
        plt.ylabel(r'Field amplitude $b$ (T)')

        # plt.annotate(r'Coupling', (np.log10(1.0), np.log10(0.007)), color='white')
        # plt.annotate(r'Filament', (np.log10(0.012), np.log10(0.74)), color='white')
        # plt.annotate(r'(uncoupled)', (np.log10(0.012), np.log10(0.55)), color='white')
        # plt.annotate(r'Filament', (np.log10(45), np.log10(0.38)), color='white')
        # plt.annotate(r'(coupled)', (np.log10(45), np.log10(0.28)), color='white')
        # plt.annotate(r'Eddy', (np.log10(2000), np.log10(0.03)), color='white')

        # ax.plot(np.log10(0.03), np.log10(0.2),  'o', color='white')#, xytext=(np.log10(0.03), np.log10(0.12)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(30),   np.log10(1),    'o', color='white')#, xytext=(np.log10(40), np.log10(0.8)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(3),    np.log10(0.2),  'o', color='white')#, xytext=(np.log10(2), np.log10(0.2)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(5000), np.log10(0.2),  'o', color='white')#, xytext=(np.log10(5000), np.log10(0.1)), arrowprops=dict(facecolor='black', shrink=0.02))

        # ax.annotate('(a)', xy=(np.log10(0.03), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(b)', xy=(np.log10(3), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(c)', xy=(np.log10(30), np.log10(1)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(d)', xy=(np.log10(5000), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')

        # Define custom tick labels for x-axis
        x_min_log = int(np.log10(min([eval('sd.' + lm.x_val) for sd in self.simulation_collection])))
        x_max_log = int(np.log10(max([eval('sd.' + lm.x_val) for sd in self.simulation_collection])))
        x = np.arange(x_min_log, x_max_log + 1)
        # Create a list of minor ticks
        minor_x_labels = []
        # 1) Add the ticks from x_min_log to ceil(x_min_log) to the minor_x_test list
        new_ticks = np.linspace(10.0 ** np.floor(x_min_log), 10.0 ** np.ceil(x_min_log), 10)[:-1]
        new_ticks = np.unique(new_ticks[new_ticks >= 10.0 ** x_min_log])
        minor_x_labels.extend(new_ticks)
        # 2) Add the ticks from ceil(x_min_log) to floor(x_max_log) to the minor_x_test list
        for x_val in x:
            new_ticks = np.linspace(10.0 ** x_val, 10.0 ** (x_val + 1), 10)[1:-1]
            if x_val == x[-1]:
                new_ticks = new_ticks[new_ticks <= 10.0 ** x_max_log]
            minor_x_labels.extend(new_ticks)
        minor_x = [np.log10(val) for val in minor_x_labels]

        new_x_labels = [f"$10^{{{val}}}$" for val in x]
        plt.xticks(x, new_x_labels)
        plt.xticks(minor_x, minor=True)

        # Define custom tick labels for y-axis
        y_min_log = np.log10(min([eval('sd.' + lm.y_val) for sd in self.simulation_collection]))
        y_max_log = np.log10(max([eval('sd.' + lm.y_val) for sd in self.simulation_collection]))
        y = np.arange(np.ceil(y_min_log), np.floor(y_max_log) + 1)
        # Create a list of minor ticks
        minor_y_labels = []
        # 1) Add the ticks from y_min_log to ceil(y_min_log) to the minor_y_test list
        new_ticks = np.linspace(10.0 ** np.floor(y_min_log), 10.0 ** np.ceil(y_min_log), 10)[:-1]
        new_ticks = np.unique(new_ticks[new_ticks >= 10.0 ** y_min_log])
        minor_y_labels.extend(new_ticks)
        # 2) Add the ticks from ceil(y_min_log) to floor(y_max_log) to the minor_y_test list
        for y_val in y:
            new_ticks = np.linspace(10.0 ** y_val, 10.0 ** (y_val + 1), 10)[1:-1]
            if y_val == y[-1]:
                new_ticks = new_ticks[new_ticks <= 10.0 ** y_max_log]
            minor_y_labels.extend(new_ticks)

        new_y_labels = [f"$10^{{{int(val)}}}$" for val in y]
        minor_y = [np.log10(val) for val in minor_y_labels]
        plt.yticks(y, new_y_labels)
        plt.yticks(minor_y, minor=True)

        # plt.savefig('C:/Users/jdular/cernbox/Documents/Reports/CERN_Reports/linkedFluxPaper/fig/loss_map_54fil_noI.pdf', bbox_inches='tight')

        if lm.save_plot:
            filePath = self.create_non_overwriting_filepath(self.outputs_folder_path, lm.filename, '.pdf', self.fdm.run.overwrite)
            plt.savefig(filePath, bbox_inches='tight')

        if self.fdm.run.launch_gui: plt.show()

    def create_errorMap(self, flux_model, display=True, iteration=None):
        """ This function creates a error map for a given rohf model on the space spanned by the simulations (f, I_amp) """

        error_type = self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.error_type

        N = len(self.simulation_collection)
        frequencies = np.zeros(N)
        peakCurrents = np.zeros(N)
        pc_error = np.zeros(N)
        dyn_error = np.zeros(N)
        flux_error = np.zeros(N)
        for idx, simulation in enumerate(self.simulation_collection):
            frequencies[idx] = simulation.f
            peakCurrents[idx] = simulation.I_amp
            # compute rohf metrics
            rohf_flux, rohf_dyn_loss, _, rohf_pc_loss, _ = flux_model.compute(simulation.time, simulation.I_transport, sim=simulation)
            # (relative) per cycle loss error
            pc_error[idx] = abs(rohf_pc_loss - simulation.total_power_per_cycle['FilamentLoss']) / simulation.total_power_per_cycle['FilamentLoss']
            # (relative) mean dynamic error
            dyn_error[idx] = np.mean(abs((rohf_dyn_loss - simulation.power['FilamentLoss']) / simulation.power['FilamentLoss']))
            # (relative) mean flux error
            non_zero_idx = np.nonzero(simulation.flux_internal)
            flux_error[idx] = np.mean(abs((rohf_flux[non_zero_idx] - simulation.flux_internal[non_zero_idx]) / simulation.flux_internal[non_zero_idx]))

        if error_type == 'pc_loss':
            error = pc_error
        elif error_type == 'flux':
            error = flux_error
        elif error_type == 'dyn_loss':
            error = dyn_error
        else:
            raise ValueError('Unrecognized error type')

        # plot errorMap result
        if display:

            show_descriptions = False
            freq_i, curr_i = np.linspace(min(frequencies), max(frequencies), 1000), np.linspace(min(peakCurrents), max(peakCurrents), 1000)
            freq_i, curr_i = np.meshgrid(freq_i, curr_i)
            plt.figure()
            ax = plt.gca()
            title = str(error_type) + " map"
            if show_descriptions:
                ax.set_title(title)
                ax.set_xlabel(r'Frequency $f$ ($\mathrm{Hz}$)')
                ax.set_ylabel(r"Transport current ratio $I/I_\mathrm{c}$")
            else:
                ax.set_axis_off()

            if self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.interpolate_error_map:
                rbf = interpolate.Rbf(frequencies, peakCurrents, error, function='linear')
                error_i = rbf(freq_i, curr_i)
                plt.imshow(error_i, vmin=error.min(), vmax=error.max(), origin='lower', aspect='auto',
                           extent=[min(frequencies), max(frequencies), min(peakCurrents) / max(peakCurrents), 1])
            plt.scatter(frequencies, peakCurrents / max(peakCurrents), c=error, edgecolor='w')
            ax.set_xscale('log')
            # ax.set_yscale('log')
            plt.colorbar()
            plt.show()
        else:
            # cumulative error over map
            return sum(error)

    def plot2d(self):
        """
        This function is used to create a 2d plot. It is supposed to be flexible and work for various kinds of plots one may want to create.
        """

        def apply_plot_settings():
            if ref:
                plt.plot(ref_array[0], ref_array[1], linestyle='--', label=self.fdm.magnet.postproc.batch_postproc.plot2d.reference_label)
            if self.fdm.magnet.postproc.batch_postproc.plot2d.legend:
                plt.legend()
            plt.grid(linestyle='dotted')
            plt.xlabel(self.fdm.magnet.postproc.batch_postproc.plot2d.xlabel)
            plt.ylabel(self.fdm.magnet.postproc.batch_postproc.plot2d.ylabel)
            plt.title(title)
            if self.fdm.magnet.postproc.batch_postproc.plot2d.x_log:
                plt.xscale('log')
            if self.fdm.magnet.postproc.batch_postproc.plot2d.y_log:
                plt.yscale('log')

        if self.tex_installation:
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            # plt.rcParams['font.size'] = 20

        plotdata_list = []
        # use labelcount to subdevide simulations into sets of equal length
        labellen = int(len(self.fdm.magnet.postproc.batch_postproc.plot2d.labels))
        grplen = int(len(self.simulation_collection) / labellen)

        # Create the title (or titles if combined_plot is False)
        title = self.fdm.magnet.postproc.batch_postproc.plot2d.title
        if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot:
            sd = self.simulation_collection[0]
            commands = re.findall('<<(.*?)>>', title)
            for c in commands:
                title = title.replace(f"<<{c}>>", str(eval('sd.' + c)))
        else:
            titles = []
            for sd in self.simulation_collection:
                commands = re.findall('<<(.*?)>>', title)
                title_i = title
                for c in commands:
                    title_i = title_i.replace(f"<<{c}>>", str(eval('sd.' + c)))
                titles.append(title_i)

        # colors = plt.cm.get_cmap('magma').resampled(len(self.simulation_collection)).colors
        max_len = max([len(self.simulation_collection), len(self.fluxmodel_collection)]) if self.fluxmodel_collection else len(self.simulation_collection)
        colors = plt.cm.get_cmap('viridis').resampled(max_len).colors

        # convert selected simulation parameters into list of PlotData
        simulation_idx = []  # nested list with indices
        fluxmodel_idx = []
        style_cycle = ['--', ':', '-.']
        for i, simulation in enumerate(self.simulation_collection):
            # get x_data
            x_data = eval(self.fdm.magnet.postproc.batch_postproc.plot2d.x_val)
            # base label
            if len(self.fdm.magnet.postproc.batch_postproc.plot2d.labels) == len(self.simulation_collection):
                label_idx = i
            else:
                label_idx = 0
            label_cmd = re.findall('<<(.*?)>>', self.fdm.magnet.postproc.batch_postproc.plot2d.labels[label_idx])
            sim_label = self.fdm.magnet.postproc.batch_postproc.plot2d.labels[label_idx].replace(f"<<{label_cmd}>>", str(eval('simulation.' + label_cmd))) if label_cmd else self.fdm.magnet.postproc.batch_postproc.plot2d.labels[label_idx]
            # add (multiple) y_data sets w. label
            for j in range(len(self.fdm.magnet.postproc.batch_postproc.plot2d.y_vals)):
                y_data = eval(self.fdm.magnet.postproc.batch_postproc.plot2d.y_vals[j])

                sublabel = sim_label if len(self.fdm.magnet.postproc.batch_postproc.plot2d.y_vals) == 1 else sim_label + ' (' + str(i) + ')'
                plotdata_list.append(PlotData(x_data=x_data, y_data=y_data, label=sublabel, color=colors[i], linestyle='-'))
                simulation_idx.append(len(plotdata_list) - 1)
            # add multiple fluxmodel results
            if self.fluxmodel_collection:
                for k, fluxmodel in enumerate(self.fluxmodel_collection):
                    y_data = eval(self.fdm.magnet.postproc.batch_postproc.plot2d.y_val_fluxmodel)
                    plotdata_list.append(PlotData(x_data=x_data, y_data=y_data, label=fluxmodel.label + '_' + str(sim_label), color=colors[i], linestyle=style_cycle[k % len(style_cycle)]))
                    fluxmodel_idx.append(len(plotdata_list) - 1)

                    # Load reference - if defined
        ref = self.fdm.magnet.postproc.batch_postproc.plot2d.reference_vals
        ref_array = []
        if ref:
            for i, ref_val in enumerate(ref):
                commands = re.findall('<<(.*?)>>', ref_val)
                for c in commands:
                    ref_val = ref_val.replace(f"<<{c}>>", str('sd.' + c))
                ref_array.append(eval(ref_val))

        # Create result folder:
        if not self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot and self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:
            # Create a folder to save the plots if combined_plot is False and save_plot is True:
            filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
            folder_path = self.create_non_overwriting_filepath(self.outputs_folder_path, filename, '', self.fdm.run.overwrite)
            if not os.path.exists(folder_path): os.makedirs(folder_path)

        # plot data as continous graphs in one plot if there are no scalars
        scalar_plotdata = any([hasattr(plotdata.x_data, "__len__") for plotdata in plotdata_list])
        if not scalar_plotdata:
            for grp_idx, grp_plotdata in enumerate([plotdata_list[i::labellen] for i in range(labellen)]):
                plt.plot([data.x_data for data in grp_plotdata], [data.y_data for data in grp_plotdata], self.fdm.magnet.postproc.batch_postproc.plot2d.linestyle, label=self.fdm.magnet.postproc.batch_postproc.plot2d.labels[grp_idx], color=colors[-grp_idx])
                if not self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot:
                    plt.figure()
            if self.fluxmodel_collection:
                for grp_idx, group in enumerate([fluxmodel_idx[i:i + grplen] for i in range(0, len(fluxmodel_idx), grplen)]):
                    plt.plot([plotdata_list[idx].x_data for idx in group], [plotdata_list[idx].y_data for idx in group], '-o', label="ROHF", color='red')
            apply_plot_settings()
        # plot data as points in one graph each in one plot
        else:
            print('test')
            for idx, plotdata in enumerate(plotdata_list):
                plt.plot(plotdata.x_data, plotdata.y_data, linestyle=plotdata.linestyle, label=plotdata.label, color=plotdata.color)
                apply_plot_settings()
        if self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:
            filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
            filePath = self.create_non_overwriting_filepath(self.outputs_folder_path, filename, '.png', self.fdm.run.overwrite)
            plt.savefig(filePath, dpi=300)

        if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot and self.fdm.magnet.postproc.batch_postproc.plot2d.save_pgfdata:
            if not scalar_plotdata:
                data_tuple = ()
                data_names = []
                for grp_idx, grp_plotdata in enumerate([plotdata_list[i::labellen] for i in range(labellen)]):
                    data_tuple = data_tuple + ([data.x_data for data in grp_plotdata], [data.y_data for data in grp_plotdata])
                    data_names.extend([self.fdm.magnet.postproc.batch_postproc.plot2d.labels[grp_idx] + str("_x"), self.fdm.magnet.postproc.batch_postproc.plot2d.labels[grp_idx] + str("_y")])
                data = np.vstack(data_tuple).T
                header = " ".join(data_names)
            else:
                max_len = max([len(plotdata.x_data) for plotdata in plotdata_list])
                # pad shorter data columns with nan - is ignored by pgfplots
                data = np.vstack(tuple([np.pad(plotdata.x_data, (0, max_len - len(plotdata.x_data)), "constant", constant_values=(np.nan,)) for plotdata in plotdata_list]
                                       + [np.pad(plotdata.y_data, (0, max_len - len(plotdata.y_data)), "constant", constant_values=(np.nan,)) for plotdata in plotdata_list])).T
                header = " ".join([plotdata.label + '_x' for plotdata in plotdata_list]) + " " + " ".join([plotdata.label + '_y' for plotdata in plotdata_list])
            np.savetxt(os.path.join(self.outputs_folder_path, 'pgfdata.txt'), data, header=header, comments="")

        if self.fdm.run.launch_gui: plt.show()



from ruamel.yaml import YAML
import argparse
import os

def update_data_settings_getdp_path(file_path, env_variable):
    # Load the YAML file
    yaml = YAML()

    with open(file_path, 'r') as file:
        config = yaml.load(file)

    # Get the value of the environment variable
    env_value = os.environ.get(env_variable)
    if env_value is None:
        raise ValueError(f"Environment variable '{env_variable}' is not set.")

    # Update the GetDP_path with the expanded value
    config['GetDP_path'] = "C:/cerngetdp/cws/getdp_" + env_value + ".exe"
    # Save the updated YAML file, preserving formatting
    with open(file_path, 'w') as file:
        yaml.dump(config, file)

    print(f"Updated 'GetDP_path' to use environment variable: {env_variable}")

if __name__ == "__main__":
    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Update GetDP_path in a YAML file to use an environment variable.")
    parser.add_argument("file", help="Path to the YAML file.")
    parser.add_argument("env_variable", help="Name of the environment variable to use for GetDP_path.")
    args = parser.parse_args()
    
    # Call the function with the provided arguments
    update_data_settings_getdp_path(args.file, args.env_variable)


from pydantic import BaseModel, Field
from typing import Union, Literal, Optional, List

# ------------------- Jc fits ---------------------------#
class ConstantJc(BaseModel):
    """
    Level 3: Class for setting constant Jc
    """

    type: Literal["Constant Jc"]
    Jc_constant: Optional[float] = None  # [A/m^2]


class Ic_A_NbTi(BaseModel):
    """
    Level 3: Class for setting IcNbTi fit
    """

    type: Literal["Ic_A_NbTi"]
    Jc_5T_4_2K: Optional[float] = None  # [A/m^2]


class Bottura(BaseModel):
    """
    Level 3: Class for setting Bottura fit
    """

    type: Literal["Bottura"]
    Tc0_Bottura: Optional[float] = None  # [K]
    Bc20_Bottura: Optional[float] = None  # [T]
    Jc_ref_Bottura: Optional[float] = None  # [A/m^2]
    C0_Bottura: Optional[float] = None  # [-]
    alpha_Bottura: Optional[float] = None  # [-]
    beta_Bottura: Optional[float] = None  # [-]
    gamma_Bottura: Optional[float] = None  # [-]


class CUDI1(BaseModel):
    """
    Level 3: Class for Nb-Ti fit based on "Fit 1" in CUDI manual
    """

    type: Literal["CUDI1"]
    Tc0_CUDI1: Optional[float] = None  # [K]
    Bc20_CUDI1: Optional[float] = None  # [T]
    C1_CUDI1: Optional[float] = None  # [A]
    C2_CUDI1: Optional[float] = None  # [A/T]


class CUDI3(BaseModel):
    """
    Level 3: Class for Nb-Ti fit based on "Fit 3" in CUDI manual
    """

    type: Literal["CUDI3"]
    Tc0_CUDI3: Optional[float] = None  # [K]
    Bc20_CUDI3: Optional[float] = None  # [T]
    c1_CUDI3: Optional[float] = None  # [-]
    c2_CUDI3: Optional[float] = None  # [-]
    c3_CUDI3: Optional[float] = None  # [-]
    c4_CUDI3: Optional[float] = None  # [-]
    c5_CUDI3: Optional[float] = None  # [-]
    c6_CUDI3: Optional[float] = None  # [-]


class Summers(BaseModel):
    """
    Level 3: Class for cable Summer's Nb3Sn fit
    """

    type: Literal["Summers"]
    Tc0_Summers: Optional[float] = None  # [K]
    Bc20_Summers: Optional[float] = None  # [T]
    Jc0_Summers: Optional[float] = None  # [A*T^0.5/m^2]


class Bordini(BaseModel):
    """
    Level 3: Class for cable Bordini's Nb3Sn fit
    """

    type: Literal["Bordini"]
    Tc0_Bordini: Optional[float] = None  # [K]
    Bc20_Bordini: Optional[float] = None  # [T]
    C0_Bordini: Optional[float] = None  # [A*T/m^2]
    alpha_Bordini: Optional[float] = None  # [-]


class Nb3Sn_HFM(BaseModel):
    """
    Level 3: Class for cable HFM Nb3Sn fit
    """

    type: Literal["Nb3Sn_HFM"]
    Tc0_Nb3Sn_HFM: Optional[float] = None  # [K]
    Bc20_Nb3Sn_HFM: Optional[float] = None  # [T]
    C0_Nb3Sn_HFM: Optional[float] = None  # [A*T/m^2]
    alpha_Nb3Sn_HFM: Optional[float] = None  # [-]
    nu_Nb3Sn_HFM: Optional[float] = None  # [-]
    p_Nb3Sn_HFM: Optional[float] = None  # [-]
    q_Nb3Sn_HFM: Optional[float] = None  # [-]


class ProDefined(BaseModel):
    """
    Level 3: Class for cable Bordini's Nb3Sn fit
    """

    type: Literal["ProDefined"]
    Tc0: Optional[float] = None  # [K]
    Bc20: Optional[float] = None  # [T]
    C0: Optional[float] = None  # [A*T/m^2]
    alpha: Optional[float] = None  # [-]
    p: Optional[float] = None  # [-]
    q: Optional[float] = None  # [-]
    v: Optional[float] = None  # [-]
    B0: Optional[float] = None  # [-]

class Succi_fixed(BaseModel):
    """
    Level 3: Class for cable Succi's YBCO fit
    """

    type: Literal["Succi_fixed"]
    Jc_factor: Optional[float] = Field(gt=0.0,
        default = 1.0,
        description = "This factor multiplies the Jc returned by the function."
    )
    # all other parameters of the Succi fit are hardcoded
class Fujikura(BaseModel):
    """
    Level 3: Class for cable Fujikura's fit
    """

    type: Literal["Fujikura"]
    Jc_factor: Optional[float] = Field(gt=0.0,
        default = 1.0,
        description = "This factor multiplies the Jc returned by the function."
    )

class BSCCO_2212_LBNL(BaseModel):
    """
    Level 3: Class for cable Bi-2212 fit developed in LBNL
    """

    # only ad-hoc fit [T. Shen, D. Davis, E. Ravaioli with LBNL, Berkeley, CA]
    type: Literal["BSCCO_2212_LBNL"]
    f_scaling_Jc_BSCCO2212: Optional[float] = None  # [-] used for the ad-hoc fit


# ------------------- Cable types ---------------------------#
class Mono(BaseModel):
    """
    Mono cable type: This is basically type of cable consisting of one strand - not really a cable
    """

    type: Literal["Mono"]
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = None
    th_insulation_along_height: Optional[float] = None
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None


class Rutherford(BaseModel):
    """
    Rutherford cable type: for example LHC MB magnet cable
    """

    type: Literal["Rutherford"]
    n_strands: Optional[int] = None
    n_strand_layers: Optional[int] = None
    n_strands_per_layers: Optional[int] = None
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = None
    th_insulation_along_height: Optional[float] = None
    width_core: Optional[float] = None
    height_core: Optional[float] = None
    strand_twist_pitch: Optional[float] = None
    strand_twist_pitch_angle: Optional[float] = None
    Rc: Optional[float] = None
    Ra: Optional[float] = None
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    f_core: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None
    material_core: Optional[str] = None
    gamma_c: Optional[float] = Field(
        default=0.0,
        description="parameter for DISCC cable homogenization"
        )


class Ribbon(BaseModel):
    """
    Mono cable type: This is basically type of cable consisting of one strand - not really a cable
    """

    type: Literal["Ribbon"]
    n_strands: Optional[int] = (
        None  # This defines the number of "strands" in the ribbon cable, which are physically glued but electrically in series
    )
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = (
        None  # This defines the thickness of the insulation around each strand (DIFFERENT FROM ROXIE CADATA FILE)
    )
    th_insulation_along_height: Optional[float] = (
        None  # This defines the thickness of the insulation around each strand (DIFFERENT FROM ROXIE CADATA FILE)
    )
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    f_core: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None
    material_core: Optional[str] = None

class TSTC(BaseModel):
    """
    Twisted Stacked-Tape Cable (TSTC) type:
    """
    type: Literal["TSTC"]
    stack_layout: Optional[List[Literal[-1,0,1]]] = Field(default=None,description="2D: Tape stack layout ordered TOP->BOTTOM. The numbers represent: 1 = a CC tape, -1 = a flipped CC tape, 0 = a shunt.")
    nb_tapes: Optional[int] = Field(default=None, description="3D: Number of tapes in the stack")
    tape_width: Optional[float] = Field(default=None, description="3D and 2D: Width of each tape")
    tape_thickness: Optional[float] = Field(default=None, description="3D and 2D: Thickness of each tape")
    twist_pitch: Optional[float] = Field(default=None, description="3D: Length over which tapes are twisted by full rotation")
    pitch_fraction: Optional[float] = Field(default=1.0, description="3D: Fraction of the twist pitch to be modelled (1.0 = full pitch, 0.5 = half pitch, etc.)")
    bare_cable_width: Optional[float] = Field(default=None, description="Cable width, typically the same as CC width")
    bare_cable_height_low: Optional[float] = Field(default=None, description="Narrow end (if applicable) cable height (thickness), typically the same as (CC thickness + Cu stabilizer thickness) * number of tapes.")
    bare_cable_height_high: Optional[float] = Field(default=None, description="Wide end (if applicable) cable height (thickness), typically the same as (CC thickness + Cu stabilizer thickness) * number of tapes.")
    bare_cable_height_mean: Optional[float] = Field(default=None, description="Average (if applicable) cable height (thickness), typically the same as (CC thickness + Cu stabilizer thickness) * number of tapes.")
    th_insulation_along_width: Optional[float] = Field(default=None, description="Insulation thickness along the width ")
    th_insulation_along_height: Optional[float] = Field(default=None, description="Insulation thickness along the height ")
    f_superconductor: Optional[float] = Field(default=None, description="Fraction of superconductor related to the total area of the cable (winding cell)")
    f_stabilizer: Optional[float] = Field(default=None, description="Fraction of stabilizer related to the total area of the cable (winding cell)")
    f_silver: Optional[float] = Field(default=None, description="Fraction of silver related to the total area of the cable (winding cell)")
    f_substrate: Optional[float] = Field(default=None, description="Fraction of substrate (including buffer layers and silver overlay) related to the total area of the cable (winding cell)")
    f_shunt: Optional[float] = Field(default=None, description="Fraction of substrate (including buffer layers and silver overlay) related to the total area of the cable (winding cell)")
    f_insulation: Optional[float] = Field(default=None, description="Fraction of cable insulation related to the total area of the cable (winding cell)")
    f_inner_voids: Optional[float] = Field(default=None, description="Fraction of additional material (typically insulation) related to the total area of the cable (winding cell)")
    f_outer_voids: Optional[float] = Field(default=None, description="Fraction of additional material (typically helium impregnating the windings) related to the total area of the cable (winding cell)")

# ------------------- Conductors ---------------------------#

# class MaterialSuperconductor(BaseModel):
#     """
#     Level 3: Class for strand superconductor material parameters
#     """
#     material: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. NbTi, Nb3Sn, etc.")
#     n_value: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
#     ec: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
#     Cv_material: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")

# class MaterialStabilizer(BaseModel):
#     """
#     Level 3: Class for strand stabilizer material parameters
#     """

#     rho_material: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
#     RRR: Optional[float] = Field(default=None, description="Residual resistivity ratio of the stabilizer.")
#     T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
#     T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
#     k_material: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
#     Cv_material: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")


class Round(BaseModel):
    """
    Level 2: Class for strand parameters
    """

    type: Literal["Round"]
    fil_twist_pitch: Optional[float] = None # Strand twist pitch
    diameter: Optional[float] = None  # ds_inGroup (LEDET), DConductor (BBQ), DStrand (ProteCCT)
    diameter_core:  Optional[float] = None  # dcore_inGroup (LEDET)
    diameter_filamentary: Optional[float] = None  # dfilamentary_inGroup (LEDET)
    filament_diameter: Optional[float] = None  # df_inGroup (LEDET)
    filament_hole_diameter: Optional[float] = Field(default=None, description="Specifies round or hexagonal hole diameter inside the filament. If None or 0.0, no hole is created.")
    number_of_filaments: Optional[int] = None  # nf_inGroup (LEDET)
    f_Rho_effective: Optional[float] = None
    Cu_noCu_in_strand: Optional[float] = None

    # -- Superconductor parameters -- #
    material_superconductor: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. Nb-Ti, Nb3Sn, etc.")
    n_value_superconductor: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
    ec_superconductor: Optional[float] = Field(default=None, description="Critical electric field of the superconductor in V/m.")
    minimum_jc_fraction: Optional[float] = Field(gt=0, le=1, default=None, description="Fraction of Jc(minimum_jc_field, T) to use as minimum Jc for the power law fit to avoid division by zero when Jc(B_local, T) decreases to zero."
                                                                           "Typical value would be 0.001 (so the Jc_minimum is 0.1% of Jc(minimum_jc_field, T))"
                                                                            "This fraction is only allowed to be greater than 0.0 and less than or equal to 1.0")
    minimum_jc_field: Optional[float] = Field(default=None, description="Magnetic flux density in tesla used for calculation of Jc(minimum_jc_field, T). This gets multiplied by minimum_jc_fraction and used as minimum Jc for the power law")
    k_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the superconductor.")
    Cv_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")
    # -- Stabilizer parameters -- #
    #material_stabilizer: Optional[str] = Field(default=None, description="Material of the stabilizer.")
    rho_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
    rho_material_holes: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the holes in the filaments."
                                                                                     "Constant resistivity can be given as float, material name as a string or None or 0.0 to use 'air' in the holes.")
    RRR: Optional[Union[float, List[float]]] = Field(default=None, description="Residual resistivity ratio of the stabilizer. If a list of RRR is provided it needs to match in length the number of matrix regions in the geometry (typically 3)")
    T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
    T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
    k_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
    Cv_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")

    # superconductor: MaterialSuperconductor = MaterialSuperconductor()
    # stabilizer: MaterialStabilizer = MaterialStabilizer()


class Rectangular(BaseModel):
    """
    Level 2: Class for strand parameters
    """

    type: Literal["Rectangular"]
    bare_width: Optional[float] = None
    bare_height: Optional[float] = None
    Cu_noCu_in_strand: Optional[float] = None
    filament_diameter: Optional[float] = None  # df_inGroup (LEDET)
    f_Rho_effective: Optional[float] = None
    fil_twist_pitch: Optional[float] = None
    bare_corner_radius: Optional[float] = None

    # -- Superconductor parameters -- #
    material_superconductor: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. NbTi, Nb3Sn, etc.")
    n_value_superconductor: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
    ec_superconductor: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
    minimum_jc_fraction: Optional[float] = Field(gt=0, le=1, default=None, description="Fraction of Jc(minimum_jc_field, T) to use as minimum Jc for the power law"
                                                                                       " fit to avoid division by zero when Jc(B_local, T) decreases to zero."
                                                                           "Typical value would be 0.001 (so the Jc_minimum is 0.1% of Jc(minimum_jc_field, T))"
                                                                            "This fraction is only allowed to be greater than 0.0 and less than or equal to 1.0")
    minimum_jc_field: Optional[float] = Field(default=None, description="Magnetic flux density in tesla used for calculation of Jc(minimum_jc_field, T)."
                                                                        "This gets multiplied by minimum_jc_fraction and used as minimum Jc for the power law")
    k_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the superconductor.")
    Cv_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")
    # -- Stabilizer parameters -- #
    #material_stabilizer: Optional[str] = Field(default=None, description="Material of the stabilizer.") #TODO this should be removed as is substituted by rho, k and Cv
    k_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
    Cv_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")
    rho_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
    RRR: Optional[Union[float, List[float]]] = Field(default=None, description="Residual resistivity ratio of the stabilizer. If a list of RRR is provided it needs to match in length the number of matrix regions in the geometry (typically 3)")
    T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
    T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")

class Copper_thickness(BaseModel):
    left: Optional[float] = Field(default=None, description="On the left side.")
    right: Optional[float] = Field(default=None, description="On the right side.")
    top: Optional[float] = Field(default=None, description="On the top side.")
    bottom: Optional[float] = Field(default=None, description="On the bottom side.")

class Silver_thickness(BaseModel):
    top: Optional[float] = Field(default=None, description="On the top side.")
    bottom: Optional[float] = Field(default=None, description="On the bottom side.")

class CC(BaseModel):
    """
    Level 2: Class for coated conductor parameters
    """
    type: Literal["CC"]

    # Core layer sizes
    HTS_thickness: Optional[float] = Field(default=None, description="HTS thickness in meters.")
    HTS_width: Optional[float] = Field(default=None, description="HTS width in meters.")
    number_of_filaments: Optional[int] = Field(default=1, ge=1, description="Number of HTS filaments. If 1, no striation case")
    gap_between_filaments: Optional[float] = Field(default=None, description="Gap between HTS filaments in meters. Only applies when number_of_filaments > 1.")
    substrate_thickness: Optional[float] = Field(default=None, description="Substrate layer thickness in meters.")

    # Plating/stabilizer
    copper_thickness: Copper_thickness = Field(default=Copper_thickness(), description="Copper thickness in meters")
    silver_thickness: Silver_thickness = Field(default=Silver_thickness(), description="Silver thickness in meters")

    # -- Superconductor parameters -- #
    material_superconductor: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. NbTi, Nb3Sn, etc.")
    n_value_superconductor: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
    ec_superconductor: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
    minimum_jc_fraction: Optional[float] = Field(gt=0, le=1, default=None, description="Fraction of Jc(minimum_jc_field, T) to use as minimum Jc for the power law"
                                                                                       " fit to avoid division by zero when Jc(B_local, T) decreases to zero."
                                                                           "Typical value would be 0.001 (so the Jc_minimum is 0.1% of Jc(minimum_jc_field, T))"
                                                                            "This fraction is only allowed to be greater than 0.0 and less than or equal to 1.0")
    minimum_jc_field: Optional[float] = Field(default=None, description="Magnetic flux density in tesla used for calculation of Jc(minimum_jc_field, T)."
                                                                        "This gets multiplied by minimum_jc_fraction and used as minimum Jc for the power law")
    k_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the superconductor.")
    Cv_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")
    k_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer, typically copper.")
    Cv_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer, typically copper.")
    rho_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
    RRR: Optional[Union[float, List[float]]] = Field(default=None, description="Residual resistivity ratio of the stabilizer. If a list of RRR is provided it needs to match in length the number of matrix regions in the geometry (typically 3)")
    T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
    T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
    k_material_silver: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the silver")
    Cv_material_silver: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the silver")
    rho_material_silver: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the silver. Constant resistivity can be given as float.")
    RRR_silver: Optional[Union[float, List[float]]] = Field(default=None, description="Residual resistivity ratio of the silver. If a list of RRR is provided it needs to match in length the number of matrix regions in the geometry (typically 3)")
    T_ref_RRR_high_silver: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements for silver.")
    T_ref_RRR_low_silver: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements for silver.")
    rho_material_substrate: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the substrate. Constant resistivity can be given as float.")
    k_material_substrate: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the substrate.")
    Cv_material_substrate: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the substrate.")

class Homogenized(BaseModel):
    """
    Level 2: Class for homogenized strand parameters, to be used in the Rutherford cable model
    """
    type: Literal["Homogenized"]

    # Strand diameter (used in the geometry step)
    diameter: Optional[float] = Field(default=None, description="Undeformed round strand diameter. Used in the geometry step if keep_strand_area==true, the strand is deformed while preserving its surface area. Not used otherwise.")


# ------------------- Conductors ---------------------------#


class Conductor(BaseModel):
    """
    Level 1: Class for conductor parameters
    """

    version: Optional[str] = None
    case: Optional[str] = None
    state: Optional[str] = None
    cable: Union[Rutherford, Mono, Ribbon, TSTC] = {
        "type": "Rutherford"
    }  # TODO: Busbar, Rope, Roebel, CORC, TSTC, CICC
    strand: Union[Round, Rectangular, CC, Homogenized] = {"type": "Round"}  # TODO: Tape, WIC
    Jc_fit: Union[ConstantJc, Bottura, CUDI1, CUDI3, Summers, Bordini, Nb3Sn_HFM, BSCCO_2212_LBNL, Ic_A_NbTi, ProDefined, Succi_fixed, Fujikura] = {
        "type": "CUDI1"
    }  # TODO: CUDI other numbers? , Roxie?



import sys
import os
import getpass
import platform
import shutil
import logging
import re

import numpy as np
from pathlib import Path
from time import sleep
import multiprocessing

import pandas as pd
import ruamel.yaml
import json
import gmsh
from pydantic import BaseModel

from fiqus.data.DataSettings import DataSettings
from fiqus.data.DataFiQuS import FDM

logger = logging.getLogger('FiQuS')


class LoggingFormatter(logging.Formatter):
    """
    Logging formatter class
    """
    grey = "\x1b[38;20m"  # debug level
    white = "\x1b[37;20m"  # info level
    yellow = "\x1b[33;20m"  # warning level
    red = "\x1b[31;20m"  # error level
    bold_red = "\x1b[31;1m"  # critical level

    reset = "\x1b[0m"
    format = '%(asctime)s | %(levelname)s | %(message)s'

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: white + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class FilesAndFolders:
    @staticmethod
    def read_data_from_yaml(full_file_path, data_class):
        with open(full_file_path, 'r') as stream:
            yaml = ruamel.yaml.YAML(typ='safe', pure=True)
            yaml_str = yaml.load(stream)
            if "magnet" in yaml_str:
                yaml_str["magnet"]["input_file_path"] = str(full_file_path)

        return data_class(**yaml_str)

    @staticmethod
    def write_data_to_yaml(full_file_path, dict_of_data_class, list_exceptions=[]):
        def my_represent_none(self, data):
            """
            Change data representation from empty string to "null" string
            """
            return self.represent_scalar('tag:yaml.org,2002:null', 'null')

        def flist(x):
            """
            Define a commented sequence to allow writing a list in a single row
            """
            retval = ruamel.yaml.comments.CommentedSeq(x)
            retval.fa.set_flow_style()  # fa -> format attribute
            return retval

        def list_single_row_recursively(data_dict: dict, exceptions: list):
            """
            Write lists in a single row
            :param data_dict: Dictionary to edit
            :param exceptions: List of strings defining keys that will not be written
            in a single row
            :return:
            """
            for key, value in data_dict.items():
                if isinstance(value, list) and (key not in exceptions):
                    data_dict[key] = flist(value)
                elif isinstance(value, np.ndarray):
                    data_dict[key] = flist(value.tolist())
                elif isinstance(value, dict):
                    data_dict[key] = list_single_row_recursively(value, exceptions)

            return data_dict

        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        yaml.emitter.alt_null = 'Null'
        yaml.representer.add_representer(type(None), my_represent_none)
        dict_of_data_class = list_single_row_recursively(dict_of_data_class, exceptions=list_exceptions)
        with open(full_file_path, 'w') as yaml_file:
            yaml.dump(dict_of_data_class, yaml_file)

    @staticmethod
    def write_data_model_to_yaml(full_file_path, data_model, with_comments=True, by_alias=True):
        if isinstance(data_model, BaseModel):
            # Set up YAML instance settings:
            yamlInstance = ruamel.yaml.YAML()

            # Convert the model_data to a ruamel.yaml object/dictionary:
            if with_comments:
                path_object = Path(full_file_path)
                # Add pydantic descriptions to the yaml file as comments:
                dummy_yaml_file_to_create_ruamel_object = (
                    path_object.resolve().parent.joinpath("dummy.yaml")
                )
                with open(dummy_yaml_file_to_create_ruamel_object, "w") as stream:
                    yamlInstance.dump(data_model.model_dump(by_alias=by_alias), stream)

                # Read the file:
                with open(dummy_yaml_file_to_create_ruamel_object, "r") as stream:
                    # Read the yaml file and store the date inside ruamel_yaml_object:
                    # ruamel_yaml_object is a special object that stores both the data and
                    # comments. Even though the data might be changed or added, the same
                    # object will be used to create the new YAML file to store the comments.
                    ruamel_yaml_object = yamlInstance.load(
                        dummy_yaml_file_to_create_ruamel_object
                    )

                os.remove(dummy_yaml_file_to_create_ruamel_object)

                def iterate_fields(model, ruamel_yaml_object):
                    for currentPydanticKey, value in model.model_fields.items():
                        if value.alias and by_alias:
                            currentDictionaryKey = value.alias
                        else:
                            currentDictionaryKey = currentPydanticKey

                        if value.description:
                            ruamel_yaml_object.yaml_add_eol_comment(
                                value.description,
                                currentDictionaryKey,
                            )

                        if hasattr(getattr(model, currentPydanticKey), "model_fields"):
                            new_ruamel_yaml_object = iterate_fields(
                                getattr(model, currentPydanticKey),
                                ruamel_yaml_object[currentDictionaryKey],
                            )

                            ruamel_yaml_object[currentDictionaryKey] = new_ruamel_yaml_object

                        elif isinstance(getattr(model, currentPydanticKey), list):
                            for i, item in enumerate(getattr(model, currentPydanticKey)):
                                if hasattr(item, "model_fields"):
                                    new_ruamel_yaml_object = iterate_fields(
                                        item,
                                        ruamel_yaml_object[currentDictionaryKey][i],
                                    )

                                    ruamel_yaml_object[currentDictionaryKey][i] = new_ruamel_yaml_object

                    return ruamel_yaml_object

                iterate_fields(data_model, ruamel_yaml_object)
                for currentPydanticKey, value in data_model.model_fields.items():
                    if value.alias and by_alias:
                        currentDictionaryKey = value.alias
                    else:
                        currentDictionaryKey = currentPydanticKey

                    if hasattr(getattr(data_model, currentPydanticKey), "model_fields"):
                        ruamel_yaml_object[currentDictionaryKey] = iterate_fields(
                            getattr(data_model, currentPydanticKey),
                            ruamel_yaml_object[currentDictionaryKey],
                        )

                data_dict = ruamel_yaml_object

            else:
                data_dict = data_model.model_dump(by_alias=by_alias)

            yamlInstance.indent(sequence=4, offset=2)
            with open(full_file_path, 'w') as yaml_file:
                yamlInstance.dump(data_dict, yaml_file)

    @staticmethod
    def prep_folder(folder_full_path, clear: bool = False):
        if clear:
            if os.path.exists(folder_full_path):
                shutil.rmtree(folder_full_path)  # delete directory
        if not os.path.exists(folder_full_path):
            os.makedirs(folder_full_path)  # make new directory

    @staticmethod
    def get_folder_path(folder_type, folder, folder_key, overwrite, required_folder):
        """
        Method for ...
        :param folder_type:
        :type folder_type:
        :param folder:
        :type folder:
        :param folder_key:
        :type folder_key:
        :param overwrite:
        :type overwrite:
        :param required_folder:
        :type required_folder:
        :return:
        :rtype:
        """
        if required_folder and not (folder_key and overwrite):
            all_dirs = [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]
            all_relevant_dirs = [x for x in all_dirs if x.startswith(f"{folder_type}_{folder_key}")]
            if f"{folder_type}_{folder_key}" in all_relevant_dirs:
                new_folder_key = f"{folder_key}_{len(all_relevant_dirs) + 1}"
                folder_key = new_folder_key

        folder_path = os.path.join(folder, folder_type + '_' + str(folder_key))
        # Disable the line below to avoid deleting the folder # TODO: add logic to control this at a higher level
        FilesAndFolders.prep_folder(folder_path, overwrite and required_folder)
        return folder_path

    @staticmethod
    def compute_folder_key(folder_type, folder, overwrite):
        # Find all the directories in the folder
        all_dirs = [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]

        # Find all the directiories that start with the folder_type (e.g. geometry, mesh, solution)
        # Then combine them into a single string with a custom seperator (se@p)
        # Seperators are used to guarantee the directories can be split later
        all_relevant_dirs = " se@p ".join([x for x in all_dirs if x.startswith(f"{folder_type}_")])
        all_relevant_dirs = f"{all_relevant_dirs} se@p "

        # Find all the integer keys in the relevant directories
        integers_in_relevant_dirs = re.findall(rf'{folder_type}_(\d+) se@p ', all_relevant_dirs)

        if integers_in_relevant_dirs is None:
            # If there are no integers in the relevant directories, set the key to 1
            folder_key = 1
        else:
            # Make a list of integers out of the integers in the relevant directories
            integers_in_relevant_dirs = [int(x) for x in integers_in_relevant_dirs]

            # Sort the integers in the relevant directories
            integers_in_relevant_dirs.sort()

            if overwrite:
                # If overwrite is true, set the key to the largest integer in the
                # so that the folder with the largest integer key is overwritten
                if len(integers_in_relevant_dirs) == 0:
                    folder_key = 1
                else:
                    folder_key = max(integers_in_relevant_dirs)
            else:
                # If overwrite is false, then find the smallest integer key that is not
                # in the list of integers in the relevant directories
                folder_key = 1
                for i in integers_in_relevant_dirs:
                    if folder_key < i:
                        break
                    folder_key += 1

        return folder_key

    @staticmethod
    def print_welcome_graphics():
        logger.info(r" _____ _  ___        ____ ")
        logger.info(r"|  ___(_)/ _ \ _   _/ ___| ")
        logger.info(r"| |_  | | | | | | | \___ \ ")
        logger.info(r"|  _| | | |_| | |_| |___) |")
        logger.info(r"|_|   |_|\__\_\\__,_|____/ ")
        logger.info("")


class CheckForExceptions:

    @staticmethod
    def check_inputs(run):  # RunFiQuS()
        # """
        # This method raises errors when geometry, mesh or solution folders inputs are incorrect. Warnings are disabled as a trial.
        # :param run: FDM.run object
        # :type run: FDM.run
        # """
        if run.type == 'start_from_yaml':
            pass
        #     if run.geometry and not run.overwrite:
        #         warnings.warn("Warning: Geometry folder is needed only if it has to be overwritten. Ignoring it...")
        #     if run.solution or run.mesh:
        #         warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        # elif run.type == 'geometry_only':
        #     if run.solution or run.mesh:
        #         warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        # elif run.type == 'geometry_and_mesh':
        #     if run.geometry and not run.overwrite:
        #         warnings.warn("Warning: Geometry folder is needed only if it has to be overwritten. Ignoring it...")
        #     if run.mesh:
        #         warnings.warn("Warning: Mesh folder is not needed. Ignoring it...")
        elif run.type == 'mesh_and_solve_with_post_process':
            if not run.geometry:
                raise Exception('Full path to Geometry not provided. '
                                'Insert options -> reference_files -> geometry.')
            # if run.mesh and not run.overwrite:
            #     warnings.warn("Warning: Mesh folder is needed only if it has to be overwritten. Ignoring it...")
            # if run.solution:
            #     warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'mesh_only':
            if not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry.')
            # if run.solution:
            #     warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'solve_with_post_process':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            # if run.solution and not run.overwrite:
            #     warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'solve_only':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            # if run.solution and not run.overwrite:
            #     warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'post_process_only':
            if not run.mesh or not run.geometry or not run.solution:
                raise Exception('Full path to Solution not provided. '
                                'Insert options -> reference_files -> geometry, mesh, and solution.')

    @staticmethod
    def check_overwrite_conditions(folder_type, folder, folder_key):
        """
        This method prints warning related to overwrite conditions settings. This is disabled as a trial.
        :param folder_type:
        :type folder_type:
        :param folder:
        :type folder:
        :param folder_key:
        :type folder_key:
        """
        pass
        # if folder_key:
        #     if not os.path.exists(os.path.join(folder, folder_type + '_' + str(folder_key))):
        #         warnings.warn(
        #             f'The folder {folder_type}_{folder_key} does not exist. Creating it...')
        # else:
        #     warnings.warn(
        #         f'Reference number of the folder {folder_type} not provided. '
        #         f'Overwriting the latest {folder_type} folder...')


class GeometricFunctions:

    @staticmethod
    def sig_dig(n, precision=8):
        return float(np.format_float_positional(n, precision=precision))

    @staticmethod
    def points_distance(a, b):
        """
            Computes the distance between two points a and b
            :param a: list of x and y coordinates
            :param b: list of x and y coordinates
        """
        a = np.array(a)
        b = np.array(b)
        return np.linalg.norm(a - b)

    @staticmethod
    def line_through_two_points(point1, point2):
        """
            Finds coefficients of the line through two points [x1,y1] and [x2,y2]
            :param point1: 2-element list defining x/y positions of the 1st point
            :param point2: 2-element list defining x/y positions of the 2nd point
            :return: 3-element list defining the A, B, and C coefficients of the line, as in: A*x + B*y + C = 0
        """
        x1, y1 = point1[0], point1[1]
        x2, y2 = point2[0], point2[1]
        if x2 == x1:
            A = 1
            B = 0
            C = - x1
        elif y2 == y1:
            A = 0
            B = 1
            C = - y1
        else:
            A = - (y2 - y1) / (x2 - x1)
            B = + 1
            C = - (x2 * y1 - x1 * y2) / (x2 - x1)
        return [float(A), float(B), float(C)]

    @staticmethod
    def centroid(X, Y):
        """
            Computes the centroid coordinates of a non-self-intersecting closed polygon
            :param X: list of x coordinate of the vertices
            :param Y: list of y coordinate of the vertices
        """
        sum_A, sum_Cx, sum_Cy = 0, 0, 0
        for i in range(len(X)):
            index = i + 1 if i != len(X) - 1 else 0
            A = X[i] * Y[index] - X[index] * Y[i]
            sum_Cx += (X[i] + X[index]) * A
            sum_Cy += (Y[i] + Y[index]) * A
            sum_A += A
        factor = 1 / (3 * sum_A)
        return [factor * sum_Cx, factor * sum_Cy]

    @staticmethod
    def arc_center_from_3_points(a, b, c):
        """
            Computes the center coordinates of an arc passing through three points
            :param a: list of x and y coordinates of one arc point
            :param b: list of x and y coordinates of one arc point
            :param c: list of x and y coordinates of one arc point
        """
        ab = [a[0] - b[0], a[1] - b[1]]
        ac = [a[0] - c[0], a[1] - c[1]]
        sac = [a[0] * a[0] - c[0] * c[0], a[1] * a[1] - c[1] * c[1]]
        sba = [b[0] * b[0] - a[0] * a[0], b[1] * b[1] - a[1] * a[1]]
        yy = (sac[0] * ab[0] + sac[1] * ab[0] + sba[0] * ac[0] + sba[1] * ac[0]) / \
             (2 * ((c[1] - a[1]) * ab[0] - (b[1] - a[1]) * ac[0]))
        xx = (sac[0] * ab[1] + sac[1] * ab[1] + sba[0] * ac[1] + sba[1] * ac[1]) / \
             (2 * ((c[0] - a[0]) * ab[1] - (b[0] - a[0]) * ac[1]))
        return [-xx, -yy]

    @staticmethod
    def corrected_arc_center(C, pnt1, pnt2):
        """
            Computes the center coordinates of an arc from two points and a guessed center
            :param C: list of x and y coordinates of guessed center
            :param pnt1: list of x and y coordinates of first arc point
            :param pnt2: list of x and y coordinates of second arc point
        """
        if pnt1[1] < 0:
            pnt_tmp = pnt1.copy()
            pnt1 = pnt2.copy()
            pnt2 = pnt_tmp
        radius = (np.sqrt(np.square(pnt1[0] - C[0]) + np.square(pnt1[1] - C[1])) +
                  np.sqrt(np.square(pnt2[0] - C[0]) + np.square(pnt2[1] - C[1]))) / 2
        d = [0.5 * abs((pnt2[0] - pnt1[0])), 0.5 * abs((pnt1[1] - pnt2[1]))]
        aa = np.sqrt(np.square(d[0]) + np.square(d[1]))
        bb = np.sqrt(np.square(radius) - np.square(aa))
        M = [pnt1[0] + d[0]]
        if pnt2[1] < pnt1[1]:
            M.append(pnt2[1] + d[1])
            sign = [-1, -1] if pnt2[1] >= 0. else [1, 1]
        else:
            M.append(pnt1[1] + d[1])
            sign = [1, -1] if pnt2[1] >= 0. else [-1, 1]
        return [M[0] + sign[0] * bb * d[1] / aa, M[1] + sign[1] * bb * d[0] / aa]

    @staticmethod
    def arc_angle_between_point_and_abscissa(p, c):
        """
            Returns the angle of an arc with center c and endpoints at (cx + radius, cy) and (px, py)
            :param p: list of x and y coordinates of a point
            :param c: list of x and y coordinates of the arc center
        """
        theta = np.arctan2(p[1] - c[1], p[0] - c[0])
        return theta + (2 * np.pi if theta < 0 else 0)

    @staticmethod
    def intersection_between_two_lines(line1, line2):
        """
            Finds the intersection point between two lines
            :param line1: list of A, B, C (A*x + B*y + C = 0)
            :param line2: list of A, B, C (A*x + B*y + C = 0)
        """
        if line1[1] == 0.0:
            x = - line1[2] / line1[0]
            y = - (line2[0] * x + line2[2]) / line2[1]
        elif line2[1] == 0.0:
            x = - line2[2] / line2[0]
            y = - (line1[0] * x + line1[2]) / line1[1]
        else:
            a = - line1[0] / line1[1]
            c = - line1[2] / line1[1]
            b = - line2[0] / line2[1]
            d = - line2[2] / line2[1]
            x = (d - c) / (a - b)
            y = a * x + c
        return [x, y]

    @staticmethod
    def intersection_between_circle_and_line(line, circle, get_only_closest: bool = False):
        """
            Finds the intersection point/s between a circle and a line
            :param line: list of A, B, C (A*x + B*y + C = 0)
            :param circle: list of lists (x and y coordinates of the center, and point)
            :param get_only_closest: boolean to return only closest intersection point to the circle point
        """
        vertical = line[1] == 0
        c, d = circle
        r = GeometricFunctions.points_distance(c, d)
        intersect = []
        if vertical:
            m = - line[2] / line[0]
            delta = r ** 2 + 2 * m * c[0] - m ** 2 - c[0] ** 2
        else:
            m, b = - line[0] / line[1], - line[2] / line[1]
            A = m ** 2 + 1
            B = 2 * (m * b - c[0] - m * c[1])
            C = b ** 2 - r ** 2 + c[0] ** 2 + c[1] ** 2 - 2 * c[1] * b
            delta = B ** 2 - 4 * A * C

        if delta < 0:  # no intersection with the circle
            return None
        elif delta == 0:  # tangent to the circle
            x0 = m if vertical else - B / 2 / A
            y0 = c[1] if vertical else m * x0 + b
            intersect.append([x0, y0])
        else:  # two intersections with the circle
            x1 = m if vertical else (- B + np.sqrt(delta)) / 2 / A
            y1 = np.sqrt(delta) + c[1] if vertical else m * x1 + b
            x2 = m if vertical else (- B - np.sqrt(delta)) / 2 / A
            y2 = - np.sqrt(delta) + c[1] if vertical else m * x2 + b
            intersect.append([x1, y1])
            intersect.append([x2, y2])
            if get_only_closest:
                distance1 = GeometricFunctions.points_distance(d, intersect[0])
                distance2 = GeometricFunctions.points_distance(d, intersect[1])
                if distance1 > distance2:
                    intersect.pop(0)
                else:
                    intersect.pop(1)
        return intersect

    @staticmethod
    def intersection_between_arc_and_line(line, arc):
        """
            Finds the intersection point/s between an arc and a line
            :param line: list of A, B, C (A*x + B*y + C = 0)
            :param arc: list of lists (x and y coordinates of the center, high-angle endpoint, and low-angle endpoint)
        """
        vertical = line[1] == 0
        c, d, e = arc
        r = GeometricFunctions.points_distance(c, d)
        angle_d = GeometricFunctions.arc_angle_between_point_and_abscissa(d, c)
        if angle_d == 0:
            angle_d = 2 * np.pi  # if the 'high-angle' angle is 0, set it to 2*pi to avoid issues with the arc
        angle_e = GeometricFunctions.arc_angle_between_point_and_abscissa(e, c)
        intersect = []
        if vertical:
            m = - line[2] / line[0]
            delta = r ** 2 + 2 * m * c[0] - m ** 2 - c[0] ** 2
        else:
            m, b = - line[0] / line[1], - line[2] / line[1]
            A = m ** 2 + 1
            B = 2 * (m * b - c[0] - m * c[1])
            C = b ** 2 - r ** 2 + c[0] ** 2 + c[1] ** 2 - 2 * c[1] * b
            delta = B ** 2 - 4 * A * C

        if delta < 0:  # no intersection with the circle
            return None
        elif delta == 0:  # tangent to the circle
            x0 = m if vertical else - B / 2 / A
            y0 = c[1] if vertical else m * x0 + b
            angle0 = GeometricFunctions.arc_angle_between_point_and_abscissa([x0, y0], c)
            intersect0 = True if angle_e < angle0 < angle_d else False
            if intersect0:
                intersect.append([x0, y0])
            else:  # no intersection with the arc
                return None
        else:  # two intersections with the circle
            x1 = m if vertical else (- B + np.sqrt(delta)) / 2 / A
            y1 = np.sqrt(delta) + c[1] if vertical else m * x1 + b
            angle1 = GeometricFunctions.arc_angle_between_point_and_abscissa([x1, y1], c)
            intersect1 = True if (angle_e < angle1 < angle_d) or abs(angle1 - angle_e) < 1e-6 or abs(angle1 - angle_d) < 1e-6 else False
            x2 = m if vertical else (- B - np.sqrt(delta)) / 2 / A
            y2 = - np.sqrt(delta) + c[1] if vertical else m * x2 + b
            angle2 = GeometricFunctions.arc_angle_between_point_and_abscissa([x2, y2], c)
            intersect2 = True if (angle_e < angle2 < angle_d) or abs(angle2 - angle_e) < 1e-6 or abs(angle2 - angle_d) < 1e-6 else False
            if not intersect1 and not intersect2:  # no intersection with the arc
                return None
            if intersect1:  # first point intersecting the arc
                intersect.append([x1, y1])
            if intersect2:  # second point intersecting the arc
                intersect.append([x2, y2])

        return intersect


class GmshUtils:

    def __init__(self, model_name='dummy_name', verbose=True):
        self.model_name = model_name
        self.verbose = verbose

    def initialize(self, verbosity_Gmsh: int = 5):
        """
        Initialize Gmsh with options for FiQuS
        :param verbosity_Gmsh: Input file run.verbosity_Gmsh
        :type verbosity_Gmsh: int
        """
        if not gmsh.is_initialized():
            gmsh.initialize(sys.argv, interruptible=False, readConfigFiles=False)
            gmsh.model.add(str(self.model_name))
            num_threads = multiprocessing.cpu_count()
            gmsh.option.setNumber('General.NumThreads', num_threads)  # enable multithreading (this seems to be only for meshing)
            gmsh.option.setNumber('Mesh.MaxNumThreads1D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads2D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads3D', num_threads)
            gmsh.option.setNumber('Geometry.OCCParallel', 1)
            gmsh.option.setNumber('Geometry.ToleranceBoolean', 0.000001)
            gmsh.option.setString('Geometry.OCCTargetUnit', 'M')
            gmsh.option.setNumber("General.Verbosity", verbosity_Gmsh)
            if self.verbose:
                gmsh.option.setNumber('General.Terminal', 1)
            else:
                gmsh.option.setNumber('General.Terminal', 0)

    def check_for_event(self):  # pragma: no cover
        action = gmsh.onelab.getString("ONELAB/Action")
        if len(action) and action[0] == "check":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------check----------------")
            gmsh.fltk.update()
            gmsh.graphics.draw()
        if len(action) and action[0] == "compute":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------compute----------------")
            gmsh.onelab.setChanged("Gmsh", 0)
            gmsh.onelab.setChanged("GetDP", 0)
            gmsh.fltk.update()
            gmsh.graphics.draw()
        return True

    def launch_interactive_GUI(self, close_after=-1):  # pragma: no cover
        gmsh.fltk.initialize()
        while gmsh.fltk.isAvailable() and self.check_for_event():
            gmsh.fltk.wait()
            if close_after >= 0:
                sleep(close_after)
                gmsh.fltk.finalize()
        gmsh.finalize()


class RoxieParsers:
    def __init__(self, conductor, block, xyCorner):
        self.conductor = conductor
        self.block = block
        self.xyCorner = xyCorner

    @staticmethod
    def parseMap2d(map2dFile: Path, physical_quantity: str = 'magnetic_flux_density'):
        """
            Generates pandas data frame with map2d content
            :param map2dFile: path of map2dFile containing the content to parse
            :param physical_quantity: magnetic_flux_density or temperature
        """
        physical_quantities_abbreviations = {'magnetic_flux_density': ('BX/T', 'BY/T'), 'temperature': ('T/K', '-')}
        columns = ['BL.', 'COND.', 'NO.', 'X-POS/MM', 'Y-POS/MM'] + \
                  [abbr for abbr in physical_quantities_abbreviations[physical_quantity]] + \
                  ['AREA/MM**2', 'CURRENT', 'FILL FAC.']
        return pd.read_csv(map2dFile, sep=r"\s{2,}|(?<=2) |(?<=T) ", engine='python', usecols=columns)

    @staticmethod
    def parseCond2d(cond2dFile: Path):
        """
            Read input file and return list of ConductorPosition objects

            # input: fileName
            # output: conductorPositionsList

        """
        # conductorStartKeyword = "CONDUCTOR POSITION IN THE CROSS-SECTION"
        blockStartKeyword = "BLOCK POSITION IN THE CROSS-SECTION"

        fileContent = open(cond2dFile, "r").read()

        # separate rows
        fileContentByRow = fileContent.split("\n")

        # Find block definition
        for i in range(len(fileContentByRow)):
            if blockStartKeyword in fileContentByRow[i]:
                startOfBlockDefinitionIndex = i

        # separate part of the data with conductor position information
        conductorPositions = fileContentByRow[5:startOfBlockDefinitionIndex - 2]

        # drop every 5th row
        conductorPositionsFourVertices = list(conductorPositions)
        del conductorPositionsFourVertices[4::5]

        # arrange data in a list of lists
        outputConductorPositions = []
        for row in conductorPositionsFourVertices:
            rowSplitStr = row.split(',')
            rowSplitFloat = [float(elem) for elem in rowSplitStr]
            outputConductorPositions.append(rowSplitFloat)

        # arrange data from list to numpy.array
        outputConductorPositionsMatrix = np.array(outputConductorPositions)

        # input: outputConductorPositions
        # output: conductorPositionsList
        conductorPositionsList = []
        for i in range(0, len(outputConductorPositions), 4):
            out = outputConductorPositions[i]
            conductor = int(out[1])
            block = int(out[2])
            xyCorner = outputConductorPositionsMatrix[i:i + 4, 4:6]
            conductorPositionsList.append(RoxieParsers(conductor, block, xyCorner))

        return conductorPositionsList


def initialize_logger(work_folder: str = None, time_stamp: str = None, verbose: bool = True, ):
    """
    This is logger function to write FiQuS log files.

    :param work_folder: Folder where the log file is written to
    :type work_folder: str
    :param time_stamp: time stamp put in the log file name
    :type time_stamp: str
    :param verbose: if true INFO level logs are printed, if false only WARNING level logs are printed to the console
    :type verbose: bool
    :return: logger object
    :rtype: object
    """

    logger = logging.getLogger('FiQuS')

    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            handler.close()
            logger.removeHandler(handler)

    if verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.INFO)
    stdout_handler.setFormatter(LoggingFormatter())
    logger.addHandler(stdout_handler)

    FilesAndFolders.prep_folder(work_folder)
    FilesAndFolders.prep_folder(os.path.join(work_folder, "logs"))
    file_handler = logging.FileHandler(os.path.join(work_folder, "logs", f"{time_stamp}.FiQuS.log"))
    file_handler.setLevel(logging.INFO)
    fileFormatter = logging.Formatter('%(asctime)s | %(levelname)s | %(message)s')
    file_handler.setFormatter(fileFormatter)
    logger.addHandler(file_handler)

    errorsAndWarnings_file_handler = logging.FileHandler(os.path.join(work_folder, "logs", f"ERRORS_WARNINGS_{time_stamp}.FiQuS.log"))
    errorsAndWarnings_file_handler.setLevel(logging.WARNING)
    errorsAndWarnings_file_handler.setFormatter(fileFormatter)
    logger.addHandler(errorsAndWarnings_file_handler)

    return logger


def create_json_schema(data_model: FDM):
    """
    Create the JSON Schema from a Pydantic data model
    :param data_model: FDM
    :type data_model: FDM
    """

    # Generate the raw JSON schema from the Pydantic model
    json_schema_dict = data_model.model_json_schema()

    # Replace anyOf with oneOf for better compatibility
    json_schema_str = json.dumps(json_schema_dict)
    json_schema_str = json_schema_str.replace("anyOf", "oneOf")

    # Pretty-print the schema with proper indentation
    pretty_json_schema = json.dumps(json.loads(json_schema_str), indent=4, ensure_ascii=False)

    # Define the output folder for the schema
    docs_folder = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "docs"
    )

    # Create the _inputs folder for the JSON schema
    json_schema_file_path = os.path.join(docs_folder, "schema.json")
    os.makedirs(os.path.dirname(json_schema_file_path), exist_ok=True)

    # Write the prettified JSON schema to a file
    with open(json_schema_file_path, "w", encoding="utf-8") as file:
        file.write(pretty_json_schema)


def get_data_settings(GetDP_path=None, settings=None):
    user_name = getpass.getuser()

    if user_name == 'root':
        user_name = 'SYSTEM'
    elif user_name == 'MP-WIN-02$':
        user_name = 'MP_WIN_02'
    if not settings:
        path_to_settings_file = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "tests", f"settings.{user_name}.yaml")
        try:
            settings = FilesAndFolders.read_data_from_yaml(path_to_settings_file, DataSettings)
        except:
            with open(settings.error.log, 'a') as file:
                # Append the string to the file
                file.write(f'Could not find: {path_to_settings_file}' + '\n')
            raise ValueError(f'File: {path_to_settings_file} does not exist.')

    if GetDP_path:
        settings.GetDP_path = GetDP_path

    return settings



import unittest
from tests.utils.fiqus_test_classes import FiQuSGeometryTests


class TestGeometryGenerators(FiQuSGeometryTests):
    def test_Pancake3D(self):
        """
        Checks if Pancake3D geometry generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_Pancake3D_REF",
            "TEST_Pancake3D_REFStructured",
            "TEST_Pancake3D_TSA",
            "TEST_Pancake3D_TSAStructured",
            "TEST_Pancake3D_TSAInsulating",
            "TEST_Pancake3D_TSAInsulatingJcVsLength"
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.geometry.N = 3

                self.generate_geometry(data_model, model_name)

                # Compare the number of entities with the reference file:
                geometry_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                self.compare_number_of_entities(geometry_file, reference_file)

                # Compare the volume information files:
                vi_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="vi"
                )
                reference_vi_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="vi"
                )
                self.compare_json_or_yaml_files(vi_file, reference_vi_file)

    def test_ConductorAC_Strand(self):
        """
        Checks if ConductorAC geometry generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_CAC_Strand_adaptiveMesh",
            "TEST_CAC_Strand_hexFilaments",
            "TEST_CAC_wireInChannel",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.geometry.N = 3

                self.generate_geometry(data_model, model_name)

                # Compare the number of entities with the reference file:
                geometry_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                self.compare_number_of_entities(geometry_file, reference_file)

                # Compare the Geometry YAML files:
                geometry_yaml_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name='GeometryModel', file_extension="yaml"
                )
                reference_geometry_yaml_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name='GeometryModel', file_extension="yaml"
                )
                self.compare_json_or_yaml_files(geometry_yaml_file, reference_geometry_yaml_file, tolerance=1e-9)

    def test_ConductorAC_Rutherford(self):
            model_names = [
                "TEST_CAC_Rutherford",
            ]

            for model_name in model_names:
                with self.subTest(model_name=model_name):
                    data_model = self.get_data_model(model_name)

                    self.generate_geometry(data_model, model_name)

                    # Compare the number of entities with the reference file:
                    geometry_file = self.get_path_to_generated_file(
                        data_model=data_model, model_name=model_name, file_extension="brep"
                    )
                    reference_file = self.get_path_to_reference_file(
                        data_model=data_model, model_name=model_name, file_extension="brep"
                    )
                    self.compare_number_of_entities(geometry_file, reference_file)

    def test_ConductorAC_CC(self):
        model_names = [
            "TEST_CAC_CC",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                self.generate_geometry(data_model, model_name)

                # Compare the number of entities with the reference file:
                geometry_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="brep"
                )
                self.compare_number_of_entities(geometry_file, reference_file)

    def test_Multipole(self):
        """
        Checks if Multipole geometry generators work correctly by comparing generated
        geometry files to reference files that were checked manually.
        """
        model_names = [
            "TEST_MULTIPOLE_MBH_1in1_TSA_withQH",
            "TEST_MULTIPOLE_MBH_1in1_TSA",
            "TEST_MULTIPOLE_MBH_1in1_REF",
            "TEST_MULTIPOLE_SMC_TSA_withQH",
            "TEST_MULTIPOLE_SMC_TSA",
            "TEST_MULTIPOLE_SMC_REF",
            "TEST_MULTIPOLE_4COND_TSA",
            "TEST_MULTIPOLE_FALCOND_C_TSA_COLLAR_POLE",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.geometry.N = 3

                self.generate_geometry(data_model, model_name)

                thermal_model_name = f"{model_name}_TH"
                electromagnetic_model_name = f"{model_name}_EM"

                # Compare the number of entities with the reference file:
                geometry_file_EM = self.get_path_to_generated_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="brep"
                )
                reference_file_EM = self.get_path_to_reference_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="brep"
                )
                self.compare_number_of_entities(geometry_file_EM, reference_file_EM)

                geometry_file_TH = self.get_path_to_generated_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="brep"
                )
                reference_file_TH = self.get_path_to_reference_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="brep"
                )
                self.compare_number_of_entities(geometry_file_TH, reference_file_TH)

                # Compare the aux files:
                aux_file_EM = self.get_path_to_generated_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="aux"
                )
                reference_aux_file_EM = self.get_path_to_reference_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="aux"
                )
                self.compare_json_or_yaml_files(aux_file_EM, reference_aux_file_EM)

                aux_file_TH = self.get_path_to_generated_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="aux"
                )
                reference_aux_file_TH = self.get_path_to_reference_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="aux"
                )
                self.compare_json_or_yaml_files(aux_file_TH, reference_aux_file_TH)

                # Compare the strs files:
                strs_file_EM = self.get_path_to_generated_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="strs"
                )
                reference_strs_file_EM = self.get_path_to_reference_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="strs"
                )
                self.compare_json_or_yaml_files(strs_file_EM, reference_strs_file_EM)

                strs_file_TH = self.get_path_to_generated_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="strs"
                )
                reference_strs_file_TH = self.get_path_to_reference_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="strs"
                )
                self.compare_json_or_yaml_files(strs_file_TH, reference_strs_file_TH)

                # Compare the crns files:
                crns_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="crns"
                )
                reference_crns_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="crns"
                )
                self.compare_json_or_yaml_files(crns_file, reference_crns_file)

                # Compare the geometry yamls:
                geometry_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name="geometry", file_extension="yaml"
                )
                reference_geometry_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name="geometry", file_extension="yaml"
                )
                self.compare_json_or_yaml_files(geometry_file, reference_geometry_file, excluded_keys="geom_file_path")

if __name__ == "__main__":
    unittest.main()
